"""
Neuronal morphology module.
This module defines classes to load and build neuronal morphologies.
"""

import abc
import numbers
import os
from abc import abstractmethod
from collections import OrderedDict, defaultdict, namedtuple

from brian2 import numpy as np
from brian2.units.allunits import meter
from brian2.units.fundamentalunits import Quantity, check_units, have_same_dimensions
from brian2.units.stdunits import um
from brian2.utils.logger import get_logger

logger = get_logger(__name__)

__all__ = ["Morphology", "Section", "Cylinder", "Soma"]


Node = namedtuple("Node", field_names="index,comp_name,x,y,z,diameter,parent,children")


def _to_meters(value):
    """
    Helper function to convert a floating point value (or array) to a `Quantity`
    in units of "meter", but also allow for ``None`` and return it as it is.
    """
    if value is None:
        return None
    else:
        return Quantity(value, dim=meter.dim)


def _from_morphology(variable, i, j):
    """
    Helper function to return coordinates from a main morphology (used by
    `SubMorphology`), dealing with ``None``.
    """
    if variable is None:
        return None
    return variable[i:j]


class MorphologyIndexWrapper:
    """
    A simpler version of `~brian2.groups.group.IndexWrapper`, not allowing for
    string indexing (`Morphology` is not a `Group`). It allows to use
    ``morphology.indices[...]`` instead of ``morphology[...]._indices()``.
    """

    def __init__(self, morphology):
        self.morphology = morphology

    def __getitem__(self, item):
        if isinstance(item, str):
            raise NotImplementedError("Morphologies do not support string indexing")
        assert isinstance(self.morphology, (SubMorphology, Morphology))
        return self.morphology._indices(item)


def _calc_start_idx(section):
    """
    Calculate the absolute start index that will be used by a flattened
    representation.
    """
    # calculate the absolute start index of this section
    # 1. find the root of the tree
    root = section
    while root._parent is not None:
        root = root._parent
    # 2. go down from the root and advance the indices until we find
    # the current section
    start_idx, found = _find_start_index(root, section)
    assert found
    return start_idx


def _find_start_index(current, target_section, index=0):
    if current == target_section:
        return index, True
    index += current.n
    for child in current.children:
        if child == target_section:
            return index, True
        else:
            index, found = _find_start_index(child, target_section, index)
            if found:
                return index, True
    return index, False


class Topology:
    """
    A representation of the topology of a `Morphology`. Has a useful string
    representation, inspired by NEURON's ``topology`` function.
    """

    def __init__(self, morphology):
        self.morphology = morphology

    def __str__(self):
        # TODO: Make sure that the shown compartments do not get out of hand
        divisor = 1
        return Topology._str_topology(self.morphology, compartments_divisor=divisor)

    @staticmethod
    def _str_topology(
        morphology, indent=0, named_path="", compartments_divisor=1, parent=None
    ):
        """
        A simple string-based representation of a morphology. Inspired by
        NEURON's ``topology`` function.
        """
        description = " " * indent
        length = max([1, morphology.n // compartments_divisor])
        if parent is not None:
            description += "`"
        if isinstance(morphology, Soma):
            description += "( )"
        else:
            description += "-" * length
            description += "|"
        if len(named_path) == 0:
            description += "  [root] \n"
        else:
            description += f"  {named_path}\n"
        for child in morphology.children:
            name = morphology.children.name(child)
            description += Topology._str_topology(
                child,
                indent=indent + 2 + length,
                named_path=f"{named_path}.{name}",
                compartments_divisor=compartments_divisor,
                parent=morphology,
            )
        return description

    __repr__ = __str__


def _rotate(vec, axis, angle):
    """
    Rotate a vector around an arbitrary axis.

    Parameters
    ----------
    vec : `ndarray`
        The vector to rotate.
    axis : `ndarray`
        The axis around which the vector should be rotated.
    angle : float
        The rotation angle (in radians).

    Returns
    -------
    rotated : `ndarray`
        The rotated vector.
    """
    return (
        vec * np.cos(angle)
        - np.cross(axis, vec) * np.sin(angle)
        + axis * np.dot(axis, vec) * (1 - np.cos(angle))
    )


def _perturb(vec, sigma):
    if sigma == 0:
        return vec
    # Get an arbitrary orthogonal vector
    if vec[1] != 0 or vec[0] != 0:
        orthogonal = np.hstack([vec[1], vec[0], 0])
    else:  # special case for the [0, 0, 1] vector
        orthogonal = np.array([1, 0, 0])

    # Rotate the orthogonal vector
    orthogonal = _rotate(orthogonal, vec, np.random.rand() * np.pi * 2)

    # Use an exponentially distributed angle for the perturbation
    perturbation = np.random.exponential(sigma, 1)
    return _rotate(vec, orthogonal, perturbation)


def _add_coordinates(
    orig_morphology,
    root=None,
    parent=None,
    name=None,
    section_randomness=0.0,
    compartment_randomness=0.0,
    n_th_child=0,
    total_children=0,
    overwrite_existing=False,
):
    # Note that in the following, all values are without physical units

    # The new direction is based on the direction of the parent section
    if parent is None:
        section_dir = np.array([0, 0, 0])
    else:
        section_dir = np.hstack(
            [
                np.asarray(parent.end_x[-1] - parent.start_x[0]),
                np.asarray(parent.end_y[-1] - parent.start_y[0]),
                np.asarray(parent.end_z[-1] - parent.start_z[0]),
            ]
        )
        parent_dir_norm = np.sqrt(np.sum(section_dir**2))
        if parent_dir_norm != 0:
            section_dir /= parent_dir_norm
        else:
            section_dir = np.array([0, 0, 0])
    if not overwrite_existing and orig_morphology.x is not None:
        section = orig_morphology.copy_section()
    elif isinstance(orig_morphology, Soma):
        # No perturbation for the soma
        section = Soma(
            diameter=orig_morphology.diameter,
            x=section_dir[0] * meter,
            y=section_dir[1] * meter,
            z=section_dir[2] * meter,
        )
    else:
        if np.sum(section_dir**2) == 0:
            # We don't have any direction to base this section on (most common
            # case is that the root section is a soma)
            # We stay in the x-y plane and distribute all children in a 360 degree
            # circle around (0, 0, 0)
            section_dir = np.array([1, 0, 0])
            rotation_axis = np.array([0, 0, 1])
            angle_increment = 2 * np.pi / total_children
            rotation_angle = np.pi / 2 + angle_increment * n_th_child
            section_dir = _rotate(section_dir, rotation_axis, rotation_angle)
        else:
            if (
                section_randomness == 0 and section_dir[2] == 0
            ):  # If we are in the x-y plane, stay there
                rotation_axis = np.array([0, 0, 1])
            else:
                rotation_axis = np.array([-section_dir[1], section_dir[2], 0])
            if section_randomness == 0:
                angle_increment = np.pi / (total_children + 1)
                rotation_angle = -np.pi / 2 + angle_increment * (n_th_child + 1)
                section_dir = _rotate(section_dir, rotation_axis, rotation_angle)
        if section_randomness > 0:
            # Rotate randomly
            section_dir = _perturb(section_dir, section_randomness)

        section_dir_norm = np.sqrt(np.sum(section_dir**2))
        section_dir /= section_dir_norm

        # For a soma, we let child sections begin at the surface of the sphere
        if isinstance(parent, Soma):
            origin = parent.diameter / 2 * section_dir
        else:
            origin = (0, 0, 0) * um
        coordinates = np.zeros((orig_morphology.n + 1, 3)) * meter
        start_coords = origin
        coordinates[0, :] = origin
        # Perturb individual compartments as well
        for idx, length in enumerate(orig_morphology.length):
            compartment_dir = _perturb(section_dir, compartment_randomness)
            compartment_dir_norm = np.sqrt(np.sum(compartment_dir**2))
            compartment_dir /= compartment_dir_norm
            current_coords = start_coords + length * compartment_dir
            coordinates[idx + 1, :] = current_coords
            start_coords = current_coords

        if isinstance(orig_morphology, Cylinder) and compartment_randomness == 0:
            section = Cylinder(
                n=orig_morphology.n,
                diameter=orig_morphology.diameter[0],
                x=coordinates[[0, -1], 0],
                y=coordinates[[0, -1], 1],
                z=coordinates[[0, -1], 2],
                type=orig_morphology.type,
            )
        elif isinstance(orig_morphology, Section):
            section = Section(
                n=orig_morphology.n,
                diameter=np.hstack(
                    [orig_morphology.start_diameter[0], orig_morphology.end_diameter]
                )
                * meter,
                x=coordinates[:, 0],
                y=coordinates[:, 1],
                z=coordinates[:, 2],
                type=orig_morphology.type,
            )
        else:
            raise NotImplementedError(
                f"Do not know how to deal with section of type {type(orig_morphology)}."
            )
    if parent is None:
        root = section
    else:
        parent.children.add(name, section)

    for idx, child in enumerate(orig_morphology.children):
        _add_coordinates(
            child,
            root=root,
            parent=section,
            name=orig_morphology.children.name(child),
            n_th_child=idx,
            total_children=len(orig_morphology.children),
            section_randomness=section_randomness,
            compartment_randomness=compartment_randomness,
            overwrite_existing=overwrite_existing,
        )
    return section


class Children:
    """
    Helper class to represent the children (sub trees) of a section. Can be
    used like a dictionary (mapping names to `Morphology` objects), but iterates
    over the values (sub trees) instead of over the keys (names).
    """

    def __init__(self, owner):
        self._owner = owner
        self._counter = 0
        self._children = []
        self._named_children = {}
        self._given_name = defaultdict(lambda: None)

    def __iter__(self):
        return iter(self._children)

    def __len__(self):
        return len(self._children)

    def __contains__(self, item):
        return item in self._named_children

    def name(self, child):
        """
        Return the given name (i.e. not the automatic name such as ``1``) for a
        child subtree.

        Parameters
        ----------
        child : `Morphology`

        Returns
        -------
        name : str
            The given name for the ``child``.
        """
        return self._given_name[child]

    def __getitem__(self, item):
        if isinstance(item, str):
            return self._named_children[item]
        else:
            raise TypeError("Index has to be an integer or a string.")

    def add(self, name, subtree, automatic_name=False):
        """
        Add a new child to the morphology.

        Parameters
        ----------
        name : str
            The name (e.g. ``"axon"``, ``"soma"``) to use for this sub tree.
        subtree : `Morphology`
            The subtree to link as a child.
        automatic_name : bool, optional
            Whether to chose a new name automatically, if a subtree of the same
            name already exists (uses e.g. ``"dend2"`` instead ``"dend"``).
            Defaults to ``False`` and will raise an error instead.
        """
        if name in self._named_children and self._named_children[name] is not subtree:
            if automatic_name:
                basename = name
                counter = 1
                while name in self._named_children:
                    counter += 1
                    name = basename + str(counter)
            else:
                raise AttributeError(f"The name {name} is already used for a subtree.")

        if subtree not in self._children:
            self._counter += 1
            self._children.append(subtree)
            self._named_children[str(self._counter)] = subtree
            self._given_name[subtree] = name

        if name is not None:
            self._named_children[name] = subtree

        subtree._parent = self._owner

    def remove(self, name):
        """
        Remove a subtree from this morphology.

        Parameters
        ----------
        name : str
            The name of the sub tree to remove.
        """
        if name not in self:
            raise AttributeError(f"The subtree {name} does not exist")
        subtree = self._named_children[name]
        del self._named_children[name]
        self._children.remove(subtree)
        subtree._parent = None

    def __repr__(self):
        n = len(self._children)
        s = f"<{int(n)} children"
        if n > 0:
            name_dict = {self.name(sec): sec for sec in self._children}
            s += f": {name_dict!r}"
        return f"{s}>"


class Morphology(metaclass=abc.ABCMeta):
    """
    Neuronal morphology (tree structure).

    The data structure is a tree where each node is an un-branched section
    consisting of a number of connected compartments, each one defined by its
    geometrical properties (length, area, diameter, position).

    Notes
    -----
    You cannot create objects of this class, create a `Soma`, a `Section`, or
    a `Cylinder` instead.
    """

    @check_units(n=1)
    def __init__(self, n, type=None):
        if isinstance(n, str):
            raise TypeError(
                "Need the number of compartments, not a string. "
                "If you want to load a morphology from a file, "
                "use 'Morphology.from_file' instead."
            )
        self._n = int(n)
        if self._n != n:
            raise TypeError("The number of compartments n has to be an integer value.")
        if n <= 0:
            raise ValueError("The number of compartments n has to be at least 1.")
        self.type = type
        self._children = Children(self)
        self._parent = None
        self.indices = MorphologyIndexWrapper(self)

    def __getitem__(self, item):
        """
        Return the subtree with the given name/index.

        Ex.: ```neuron['axon']``` or ```neuron['11213']```
        ```neuron[10*um:20*um]``` returns the subbranch from 10 um to 20 um.
        ```neuron[10*um]``` returns one compartment.
        ```neuron[5]``` returns compartment number 5.
        """
        if isinstance(item, slice):  # neuron[10*um:20*um] or neuron[1:3]
            using_lengths = all(
                [
                    arg is None or have_same_dimensions(arg, meter)
                    for arg in [item.start, item.stop]
                ]
            )
            using_ints = all(
                [
                    arg is None or int(arg) == float(arg)
                    for arg in [item.start, item.stop]
                ]
            )
            if not (using_lengths or using_ints):
                raise TypeError("Index slice has to use lengths or integers")

            if using_lengths:
                if item.step is not None:
                    raise TypeError(
                        "Cannot provide a step argument when slicing with lengths"
                    )
                pos = np.cumsum(np.asarray(self.length))  # coordinate on the section
                # We use a special handling for values very close to the points
                # between the compartments to avoid non-intuitive rounding
                # effects: a point closer than 1e-12*length of section will be
                # considered to be within the following section (for a start
                # index), respectively within the previous section (for an end
                # index)
                if item.start is None:
                    i = 0
                else:
                    diff = np.abs(float(item.start) - pos)
                    if min(diff) < 1e-12 * pos[-1]:
                        i = np.argmin(diff) + 1
                    else:
                        i = np.searchsorted(pos, item.start)
                if item.stop is None:
                    j = len(pos)
                else:
                    diff = np.abs(float(item.stop) - pos)
                    if min(diff) < 1e-12 * pos[-1]:
                        j = np.argmin(diff) + 1
                    else:
                        j = np.searchsorted(pos, item.stop) + 1
            else:  # integers
                i, j, step = item.indices(self.n)
                if step != 1:
                    raise TypeError("Can only slice a contiguous segment")
        elif isinstance(item, Quantity) and have_same_dimensions(item, meter):
            pos = np.hstack(
                [0, np.cumsum(np.asarray(self.length))]
            )  # coordinate on the section
            if float(item) < 0 or float(item) > (1 + 1e-12) * pos[-1]:
                raise IndexError(
                    f"Invalid index {item}, has to be in the interval "
                    f"[{0 * meter!s}, {pos[-1] * meter!s}]."
                )
            diff = np.abs(float(item) - pos)
            if min(diff) < 1e-12 * pos[-1]:
                i = np.argmin(diff)
            else:
                i = np.searchsorted(pos, item) - 1
            j = i + 1
        elif isinstance(item, numbers.Integral):  # int: returns one compartment
            if item < 0:  # allows e.g. to use -1 to get the last compartment
                item += self.n
            if item >= self.n:
                raise IndexError(f"Invalid index {item} for {self.n} compartments")
            i = item
            j = i + 1
        elif isinstance(item, str):
            item = str(item)  # convert int to string
            if (len(item) > 1) and all(
                [c in "LR123456789" for c in item]
            ):  # binary string of the form LLLRLR or 1213 (or mixed)
                return self._children[item[0]][item[1:]]
            elif item in self._children:
                return self._children[item]
            else:
                raise AttributeError(f"The subtree {item} does not exist")
        else:
            raise TypeError(f"Index of type {type(item)} not understood")

        return SubMorphology(self, i, j)

    def __setitem__(self, item, child):
        """
        Inserts the subtree and name it ``item``.
        Ex.: ``neuron['axon']`` or ``neuron['11213']``
        """
        item = str(item)  # convert int to string
        if (len(item) > 1) and all([c in "LR123456789" for c in item]):
            # binary string of the form LLLRLR or 1213 (or mixed)
            self.children[item[0]][item[1:]] = child
        else:
            self.children.add(item, child)

    def __delitem__(self, item):
        """
        Remove the subtree ``item``.
        """
        item = str(item)  # convert int to string
        if (len(item) > 1) and all([c in "LR123456789" for c in item]):
            # binary string of the form LLLRLR or 1213 (or mixed)
            del self._children[item[0]][item[1:]]
        else:
            self._children.remove(item)

    def __getattr__(self, item):
        """
        Return the subtree named ``item``.

        Ex.: ``axon = neuron.axon``
        """
        if item.startswith("_"):
            return super(object, self).__getattr__(item)
        else:
            return self[item]

    def __setattr__(self, item, child):
        """
        Attach a subtree and name it ``item``.

        Ex.: ``neuron.axon = Soma(diameter=10*um)``
        """
        if isinstance(child, Morphology) and not item.startswith("_"):
            self[item] = child
        else:  # If it is not a subtree, then it's a normal class attribute
            object.__setattr__(self, item, child)

    def __delattr__(self, item):
        """
        Remove the subtree ``item``.
        """
        del self[item]

    def _indices(self, item=None, index_var="_idx"):
        """
        Return compartment indices for the main section, relative to the
        original morphology.
        """
        if index_var != "_idx":
            raise AssertionError(f"Unexpected index {index_var}")
        if not (item is None or item == slice(None)):
            if isinstance(item, slice):
                # So that this always returns an array of values, even if it is
                # just a single value
                return self[item]._indices(slice(None))
            else:
                return self[item]._indices(None)
        else:
            start_idx = _calc_start_idx(self)
            if self.n == 1 and item is None:
                return start_idx
            else:
                return np.arange(start_idx, start_idx + self.n)

    def topology(self):
        """
        Return a representation of the topology

        Returns
        -------
        topology : `Topology`
            An object representing the topology (can be converted to a string
            by using ``str(...)`` or simply by printing it with `print`.)
        """
        return Topology(self)

    def generate_coordinates(
        self,
        section_randomness=0.0,
        compartment_randomness=0.0,
        overwrite_existing=False,
    ):
        r"""
        Create a new `Morphology`, with coordinates filled in place where the
        previous morphology did not have any. This is mostly useful for
        plotting a morphology, it does not affect its electrical properties.

        Parameters
        ----------
        section_randomness : float, optional
            The randomness when deciding the direction vector for each new
            section. The given number is the :math:`\beta` parameter of an
            exponential distribution (in degrees) which will be used to
            determine the deviation from the direction of the parent section.
            If the given value equals 0 (the default), then a deterministic
            algorithm will be used instead.
        compartment_randomness : float, optional
            The randomness when deciding the direction vector for each
            compartment within a section. The given number is the :math:`\beta`
            parameter of an exponential distribution (in degrees) which will be
            used to determine the deviation from the main direction of the
            current section. If the given value equals 0 (the default), then all
            compartments will be along a straight line.
        overwrite_existing : bool, optional
            Whether to overwrite existing coordinates in the morphology. This
            is by default set to ``False``, meaning that only sections that do
            not currently have any coordinates set will get new coordinates.
            This allows to conveniently generate a morphology that can be
            plotted for a morphology that is based on points but also has
            artificially added sections (the most common case: an axon added
            to a reconstructed morphology). If set to ``True``, all sections
            will get new coordinates. This can be useful to either get a
            schematic representation of the morphology (with
            ``section_randomness`` and ``compartment_randomness`` both 0) or to
            simply generate a new random variation of a morphology (which will
            still be electrically equivalent, of course).

        Returns
        -------
        morpho_with_coordinates : `Morphology`
            The same morphology, but with coordinates
        """
        # Convert to radians
        section_randomness *= np.pi / 180
        compartment_randomness *= np.pi / 180
        return _add_coordinates(
            self,
            section_randomness=section_randomness,
            compartment_randomness=compartment_randomness,
            overwrite_existing=overwrite_existing,
        )

    @abstractmethod
    def copy_section(self):
        """
        Create a copy of the current section (attributes of this section only,
        not re-creating the parent/children relation)

        Returns
        -------
        copy : `Morphology`
            A copy of this section (without the links to the parent/children)
        """
        raise NotImplementedError()

    @property
    def n(self):
        """
        The number of compartments in this section.
        """
        return self._n

    def __len__(self):
        """
        This is not well-defined, use `Morphology.n` or
        `Morphology.total_compartments` instead.
        """
        raise TypeError(
            "The 'length' of a Morphology is ambiguous, use its "
            "'n' attribute for the number of compartments in this "
            "section or the 'total_compartments' attribute for the "
            "total number of compartments in the whole sub-tree."
        )

    @property
    def total_compartments(self):
        """
        The total number of compartments in this subtree (i.e. the number of
        compartments in this section plus all the compartments in the sections
        deeper in the tree).
        """
        return self.n + sum(c.total_compartments for c in self.children)

    @property
    def total_sections(self):
        """
        The total number of sections in this subtree.
        """
        return 1 + sum(c.total_sections for c in self.children)

    @property
    def parent(self):
        """
        The parent section of this section.
        """
        return self._parent

    @property
    def children(self):
        """
        The children (as a `Children` object) of this section.
        """
        return self._children

    @property
    @abc.abstractmethod
    def end_distance(self):
        """
        The distance to the root of the morphology at the end of this section.
        """
        raise NotImplementedError()

    # Per-compartment attributes
    @property
    @abc.abstractmethod
    def area(self):
        """
        The membrane surface area of each compartment in this section.
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def volume(self):
        """
        The volume of each compartment in this section.
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def length(self):
        """
        The length of each compartment in this section.
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def r_length_1(self):
        """
        The geometry-dependent term to calculate the conductance between the
        start and the midpoint of each compartment. Dividing this value by the
        Intracellular resistivity gives the conductance.
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def r_length_2(self):
        """
        The geometry-dependent term to calculate the conductance between the
        midpoint and the end of each compartment. Dividing this value by the
        Intracellular resistivity gives the conductance.
        """
        raise NotImplementedError()

    # At-midpoint attributes
    @property
    @abc.abstractmethod
    def diameter(self):
        """
        The diameter at the middle of each compartment in this section.
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def distance(self):
        """
        The total distance between the midpoint of each compartment and the root
        of the morphology.
        """
        raise NotImplementedError()

    @property
    def start_x(self):
        """
        The x coordinate at the beginning of each compartment. Returns ``None``
        for morphologies without coordinates.
        """
        return _to_meters(self.start_x_)

    @property
    def start_y(self):
        """
        The y coordinate at the beginning of each compartment. Returns ``None``
        for morphologies without coordinates.
        """
        return _to_meters(self.start_y_)

    @property
    def start_z(self):
        """
        The z coordinate at the beginning of each compartment. Returns ``None``
        for morphologies without coordinates.
        """
        return _to_meters(self.start_z_)

    @property
    @abc.abstractmethod
    def start_x_(self):
        """
        The x coordinate (as a unitless floating point number) at the beginning
        of each compartment. Returns ``None`` for morphologies without
        coordinates.
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def start_y_(self):
        """
        The y coordinate (as a unitless floating point number) at the beginning
        of each compartment. Returns ``None`` for morphologies without
        coordinates.
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def start_z_(self):
        """
        The z coordinate (as a unitless floating point number) at the beginning
        of each compartment. Returns ``None`` for morphologies without
        coordinates.
        """
        raise NotImplementedError()

    @property
    def x(self):
        """
        The x coordinate at the midpoint of each compartment. Returns ``None``
        for morphologies without coordinates.
        """
        return _to_meters(self.x_)

    @property
    def y(self):
        """
        The y coordinate at the midpoint of each compartment. Returns ``None``
        for morphologies without coordinates.
        """
        return _to_meters(self.y_)

    @property
    def z(self):
        """
        The y coordinate at the midpoint of each compartment. Returns ``None``
        for morphologies without coordinates.
        """
        return _to_meters(self.z_)

    @property
    @abc.abstractmethod
    def x_(self):
        """
        The x coordinate (as a unitless floating point number) at the midpoint
        of each compartment. Returns ``None`` for morphologies without
        coordinates.
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def y_(self):
        """
        The y coordinate (as a unitless floating point number) at the midpoint
        of each compartment. Returns ``None`` for morphologies without
        coordinates.
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def z_(self):
        """
        The z coordinate (as a unitless floating point number) at the midpoint
        of each compartment. Returns ``None`` for morphologies without
        coordinates.
        """
        raise NotImplementedError()

    @property
    def end_x(self):
        """
        The x coordinate at the end of each compartment. Returns ``None``
        for morphologies without coordinates.
        """
        return _to_meters(self.end_x_)

    @property
    def end_y(self):
        """
        The y coordinate at the end of each compartment. Returns ``None``
        for morphologies without coordinates.
        """
        return _to_meters(self.end_y_)

    @property
    def end_z(self):
        """
        The z coordinate at the end of each compartment. Returns ``None``
        for morphologies without coordinates.
        """
        return _to_meters(self.end_z_)

    @property
    @abc.abstractmethod
    def end_x_(self):
        """
        The x coordinate (as a unitless floating point number) at the end of
        each compartment. Returns ``None`` for morphologies without coordinates.
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def end_y_(self):
        """
        The y coordinate (as a unitless floating point number) at the end of
        each compartment. Returns ``None`` for morphologies without coordinates.
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def end_z_(self):
        """
        The z coordinate (as a unitless floating point number) at the end of
        each compartment. Returns ``None`` for morphologies without coordinates.
        """
        raise NotImplementedError()

    @property
    def coordinates(self):
        r"""
        Array with all coordinates at the start- and end-points of each
        compartment in this section. The array has size :math:`(n+1) \times 3`,
        where :math:`n` is the number of compartments in this section. Each
        row is one point (start point of first compartment, end point of first
        compartment, end point of second compartment, ...), with the columns
        being the x, y, and z coordinates. Returns ``None`` for morphologies
        without coordinates.
        """
        if self.x_ is None:
            return None
        else:
            return Quantity(self.coordinates_, dim=meter.dim)

    @property
    def coordinates_(self):
        r"""
        Array with all coordinates (as unitless floating point numbers) at the
        start- and end-points of each compartment in this section. The array has
        size :math:`(n+1) \times 3`, where :math:`n` is the number of
        compartments in this section. Each row is one point (start point of
        first compartment, end point of first compartment, end point of second
        compartment, ...), with the columns being the x, y, and z coordinates.
        Returns ``None`` for morphologies without coordinates.
        """
        if self.x_ is None:
            return None
        else:
            return np.vstack(
                [
                    np.hstack([self.start_x_[0], self.end_x_[:]]),
                    np.hstack([self.start_y_[0], self.end_y_[:]]),
                    np.hstack([self.start_z_[0], self.end_z_[:]]),
                ]
            ).T

    @staticmethod
    def _create_section(compartments, name, parent, sections, spherical_soma):
        if (
            spherical_soma
            and len(compartments) == 1
            and compartments[0].comp_name == "soma"
        ):
            soma = compartments[0]
            section = Soma(
                diameter=soma.diameter * um, x=soma.x * um, y=soma.y * um, z=soma.z * um
            )
        else:
            sec_x, sec_y, sec_z, sec_diameter = zip(
                *[(c.x, c.y, c.z, c.diameter) for c in compartments]
            )
            # Add a point for the end of the parent_idx compartment
            if parent is not None:
                n = len(compartments)
                if parent.comp_name is not None and parent.comp_name.lower() == "soma":
                    # For a Soma, we don't use its diameter
                    start_diameter = sec_diameter[0]
                else:
                    start_diameter = parent.diameter
                # Use relative coordinates
                sec_x = np.array(sec_x) - parent.x
                sec_y = np.array(sec_y) - parent.y
                sec_z = np.array(sec_z) - parent.z
                start_x = start_y = start_z = 0.0
            else:
                n = len(compartments) - 1
                start_diameter = sec_diameter[0]
                sec_diameter = sec_diameter[1:]
                start_x = sec_x[0]
                start_y = sec_y[0]
                start_z = sec_z[0]
                sec_x = sec_x[1:]
                sec_y = sec_y[1:]
                sec_z = sec_z[1:]

            diameter = np.hstack([start_diameter, sec_diameter]) * um
            x = np.hstack([start_x, sec_x]) * um
            y = np.hstack([start_y, sec_y]) * um
            z = np.hstack([start_z, sec_z]) * um
            section = Section(n=n, diameter=diameter, x=x, y=y, z=z, type=name)

        # Add the section as a child to its parent
        if parent is not None:
            parent_sec = sections[parent.index]
            parent_sec.children.add(name, section, automatic_name=True)

        return section

    @staticmethod
    def _compartments_to_sections(
        compartment, spherical_soma, current_compartments=None, sections=None
    ):
        # Merge all unbranched compartments of the same type into a single
        # section
        if sections is None:
            sections = OrderedDict()
        if current_compartments is None:
            current_compartments = []

        # We have to create a new section, if we are either
        # 1. at a leaf of the tree or at a branching point, or
        # 2. if the compartment type changes
        while True:
            current_compartments.append(compartment)
            if (
                len(compartment.children) != 1
                or compartment.children[0].comp_name != compartment.comp_name
            ):
                break
            compartment = compartment.children[0]

        parent = current_compartments[0].parent
        section = Morphology._create_section(
            current_compartments,
            compartment.comp_name,
            parent=parent,
            sections=sections,
            spherical_soma=spherical_soma,
        )
        sections[current_compartments[-1].index] = section
        # If we are at a branching point, recurse into all subtrees
        for child in compartment.children:
            Morphology._compartments_to_sections(
                child,
                spherical_soma=spherical_soma,
                current_compartments=None,
                sections=sections,
            )

        return sections

    @staticmethod
    def _replace_three_point_soma(compartment, all_compartments):
        # Replace a three-point/two-cylinder soma by a single spherical soma
        # if possible (see http://neuromorpho.org/SomaFormat.html for some
        # details)

        # We are looking for a node with two children of the soma type (and
        # other childen of other types), where the two children don't have any
        # children of their own
        # To avoid recursion errors we are not searching any further
        soma_children = [c for c in compartment.children if c.comp_name == "soma"]
        if (
            compartment.comp_name == "soma"
            and len(soma_children) == 2
            and all(len(c.children) == 0 for c in soma_children)
        ):
            # This is a 3-point soma to replace
            soma_c = [compartment] + soma_children
            if not all(abs(c.diameter - soma_c[0].diameter) < 1e-15 for c in soma_c):
                indices = ", ".join(str(c.index) for c in soma_c)
                raise ValueError(
                    f"Found a '3-point-soma' (lines: {indices}), but not "
                    "all the diameters are "
                    "identical."
                )
            diameter = soma_c[0].diameter
            point_0 = np.array([soma_c[0].x, soma_c[0].y, soma_c[0].z])
            point_1 = np.array([soma_c[1].x, soma_c[1].y, soma_c[1].z])
            point_2 = np.array([soma_c[2].x, soma_c[2].y, soma_c[2].z])
            length_1 = np.sqrt(np.sum((point_1 - point_0) ** 2))
            length_2 = np.sqrt(np.sum((point_2 - point_0) ** 2))
            if (
                np.abs(length_1 - diameter / 2) > 0.1
                or np.abs(length_2 - diameter / 2) > 0.1
            ):
                raise ValueError(
                    "Cannot replace '3-point-soma' by a single "
                    "point, the second and third points should "
                    "be positioned one radius away from the "
                    f"first point. Distances are {length_1:.3f}um and "
                    f"{length_2:.3f}um, respectively, while the "
                    f"radius is {diameter / 2:.3f}um."
                )
            children = [c for c in compartment.children if c not in soma_c]
            compartment = Node(
                index=compartment.index,
                comp_name="soma",
                x=point_0[0],
                y=point_0[1],
                z=point_0[2],
                diameter=diameter,
                parent=compartment.parent,
                children=children,
            )
            all_compartments[compartment.index] = compartment
            del all_compartments[soma_children[0].index]
            del all_compartments[soma_children[1].index]
        else:
            # Raise a warning if there is more than one soma compartment
            somata = [c for c in all_compartments.values() if c.comp_name == "soma"]
            if len(somata) > 1:
                logger.warning(
                    f"Found {len(somata)} soma compartments. If you have a 3-point soma, "
                    "make sure it is at the beginning.",
                    name_suffix="soma_compartments",
                )

    @staticmethod
    def from_points(points, spherical_soma=True):
        """
        Create a morphology from a sequence of points (similar to the ``SWC``
        format, see `Morphology.from_swc_file`). Each point has to be
        a 7-tuple: ``(index, name, x, y, z, diameter, parent)``

        Note that the values should not use units, but are instead all taken
        to be in micrometers.

        Parameters
        ----------
        points : sequence of 7-tuples
            The points of the morphology.
        spherical_soma : bool, optional
            Whether to model a soma as a sphere.
        Returns
        -------
        morphology : `Morphology`

        Notes
        -----
        This format closely follows the SWC format (see
        `Morphology.from_swc_file`) with two differences: the ``type`` should
        be a string (e.g. ``'soma'``) instead of an integer and the 6-th element
        should be the diameter and not the radius.
        """
        # First pass through all points to get the dependency structure
        compartments = OrderedDict()
        for counter, point in enumerate(points):
            if len(point) != 7:
                raise ValueError(
                    "Each point needs to be described by 7 values, got"
                    f" {len(point)} instead."
                )
            index, name, x, y, z, diameter, parent_idx = point

            if index in compartments:
                raise ValueError(f"Two compartments with index {int(index)}")
            if parent_idx == index:
                raise ValueError(
                    f"Compartment {int(index)} lists itself as the parent compartment."
                )

            if counter == 0 and parent_idx == -1:
                parent = None  # The first compartment does not have a parent
            elif parent_idx not in compartments:
                raise ValueError(
                    f"Did not find the compartment {parent_idx} (parent "
                    f"compartment of compartment {index}). Make sure "
                    "that parent compartments are listed before "
                    "their children."
                )
            else:
                parent = compartments[parent_idx]
            children = []
            node = Node(index, name, x, y, z, diameter, parent, children)
            compartments[index] = node
            if parent is not None:
                parent.children.append(node)

        if spherical_soma:
            Morphology._replace_three_point_soma(
                list(compartments.values())[0], compartments
            )

        sections = Morphology._compartments_to_sections(
            list(compartments.values())[0], spherical_soma
        )

        # Go through all the sections again and add standard names for all
        # sections (potentially in addition to the name they already have):
        # "L" + "R" for one or two children, "1", "2", "3", etc. otherwise
        children_counter = defaultdict(int)
        for section in sections.values():
            parent = section.parent
            if parent is not None:
                children_counter[parent] += 1
                children = parent.children
                nth_child = children_counter[parent]
                if len(children) <= 2:
                    name = "L" if nth_child == 1 else "R"
                else:
                    name = f"{int(nth_child)}"
                children.add(name, section)

        # There should only be one section without parents
        root = [sec for sec in sections.values() if sec.parent is None]
        assert len(root) == 1
        return root[0]

    @staticmethod
    def from_swc_file(filename, spherical_soma=True):
        """
        Load a morphology from a ``SWC`` file. A large database of morphologies
        in this format can be found at http://neuromorpho.org

        The format consists of an optional header of lines starting with ``#``
        (ignored), followed by a sequence of points, each described in a line
        following the format::

            index type x y z radius parent

        ``index`` is an integer label (starting at 1) that identifies the
        current point and increases by one each line. ``type`` is an integer
        representing the type of the neural segment. The only type that changes
        the interpretation by Brian is the type ``1`` which signals a soma.
        Types ``2`` (axon), ``3`` (dendrite), and ``4`` (apical dendrite) are
        used to give corresponding names to the respective sections. All other
        types are ignored. ``x``, ``y``, and ``z`` are the cartesian coordinates
        at each point and ``r`` is its radius. ``parent`` refers to the index
        of the parent point or is ``-1`` for the root point.

        Parameters
        ----------
        filename : str
            The name of the ``SWC`` file.
        spherical_soma : bool, optional
            Whether to model the soma as a sphere.

        Returns
        -------
        morpho : `Morphology`
            The morphology stored in the given file.
        """
        swc_types = defaultdict(lambda: None)
        # The following names will be translated into names, all other will be
        # ignored
        swc_types.update({"1": "soma", "2": "axon", "3": "dend", "4": "apic"})

        with open(filename) as f:
            points = []
            for line_no, line in enumerate(f):
                line = line.strip()
                if line.startswith("#") or len(line) == 0:
                    # Ignore comments or empty lines
                    continue
                splitted = line.split()
                if len(splitted) != 7:
                    raise ValueError(
                        "Each line of an SWC file has to contain "
                        "7 space-separated entries, but line "
                        f"{line_no + 1} contains {len(splitted)}."
                    )
                index, comp_type, x, y, z, radius, parent = splitted
                points.append(
                    (
                        int(index),
                        swc_types[comp_type],
                        float(x),
                        float(y),
                        float(z),
                        2 * float(radius),
                        int(parent),
                    )
                )

        return Morphology.from_points(points, spherical_soma=spherical_soma)

    @staticmethod
    def from_file(filename, spherical_soma=True):
        """
        Convencience method to load a morphology from a given file. At the
        moment, only ``SWC`` files are supported, calling this function is
        therefore equivalent to calling `Morphology.from_swc_file` directly.

        Parameters
        ----------
        filename : str
            The name of a file storing a morphology.
        spherical_soma : bool, optional
            Whether to model the soma as a sphere.
        Returns
        -------
        morphology : `Morphology`
            The morphology stored in the given file.
        """
        _, ext = os.path.splitext(filename)
        if ext.lower() == ".swc":
            return Morphology.from_swc_file(filename, spherical_soma=spherical_soma)
        else:
            raise NotImplementedError(
                "Currently, SWC is the only supported file format."
            )


class SubMorphology:
    """
    A view on a subset of a section in a morphology.
    """

    def __init__(self, morphology, i, j):
        self._morphology = morphology
        self.indices = MorphologyIndexWrapper(self)
        self._i = i
        self._j = j

    def _indices(self, item=None):
        if not (item is None or item == slice(None)):
            raise IndexError("Cannot index a view on a subset of a section further")
        # Start index of the main section
        start_idx = _calc_start_idx(self._morphology)
        if item is None and self.n == 1:
            return start_idx + self._i
        else:
            return np.arange(start_idx + self._i, start_idx + self._j)

    @property
    def n(self):
        """
        The number of compartments in this sub-section.
        """
        return self._j - self._i

    def __len__(self):
        return self.n

    @property
    def n_sections(self):
        """
        The number of sections in this sub-section (always 1).
        """
        return 1

    # Per-compartment attributes
    @property
    def area(self):
        """
        The membrane surface area of each compartment in this sub-section.
        """
        return self._morphology.area[self._i : self._j]

    @property
    def volume(self):
        """
        The volume of each compartment in this sub-section.
        """
        return self._morphology.volume[self._i : self._j]

    @property
    def length(self):
        """
        The length of each compartment in this sub-section.
        """
        return self._morphology.length[self._i : self._j]

    @property
    def r_length_1(self):
        """
        The geometry-dependent term to calculate the conductance between the
        start and the midpoint of each compartment in this sub-section.
        Dividing this value by the Intracellular resistivity gives the
        conductance.
        """
        return self._morphology.r_length_1[self._i : self._j]

    @property
    def r_length_2(self):
        """
        The geometry-dependent term to calculate the conductance between the
        midpoint and the end of each compartment in this sub-section. Dividing
        this value by the Intracellular resistivity gives the conductance.
        """
        return self._morphology.r_length_2[self._i : self._j]

    # At-midpoint attributes
    @property
    def diameter(self):
        """
        The diameter at the middle of each compartment in this sub-section.
        """
        return self._morphology.diameter[self._i : self._j]

    @property
    def distance(self):
        """
        The total distance between the midpoint of each compartment in this
        sub-section and the root of the morphology.
        """
        return self._morphology.distance[self._i : self._j]

    @property
    def start_x(self):
        """
        The x coordinate at the beginning of each compartment in this
        sub-section. Returns ``None`` for morphologies without coordinates.
        """
        return _to_meters(self.start_x_)

    @property
    def start_y(self):
        """
        The y coordinate at the beginning of each compartment in this
        sub-section. Returns ``None`` for morphologies without coordinates.
        """
        return _to_meters(self.start_y_)

    @property
    def start_z(self):
        """
        The x coordinate at the beginning of each compartment in this
        sub-section. Returns ``None`` for morphologies without coordinates.
        """
        return _to_meters(self.start_z_)

    @property
    def start_x_(self):
        """
        The x coordinate (as a unitless floating point number) at the beginning
        of each compartment in this sub-section. Returns ``None`` for
        morphologies without coordinates.
        """
        return _from_morphology(self._morphology.start_x_, self._i, self._j)

    @property
    def start_y_(self):
        """
        The y coordinate (as a unitless floating point number) at the beginning
        of each compartment in this sub-section. Returns ``None`` for
        morphologies without coordinates.
        """
        return _from_morphology(self._morphology.start_y_, self._i, self._j)

    @property
    def start_z_(self):
        """
        The z coordinate (as a unitless floating point number) at the beginning
        of each compartment in this sub-section. Returns ``None`` for
        morphologies without coordinates.
        """
        return _from_morphology(self._morphology.start_z_, self._i, self._j)

    @property
    def x(self):
        """
        The x coordinate at the midpoint of each compartment in this
        sub-section. Returns ``None`` for morphologies without coordinates.
        """
        return _to_meters(self.x_)

    @property
    def y(self):
        """
        The y coordinate at the midpoint of each compartment in this
        sub-section. Returns ``None`` for morphologies without coordinates.
        """
        return _to_meters(self.y_)

    @property
    def z(self):
        """
        The z coordinate at the midpoint of each compartment in this
        sub-section. Returns ``None`` for morphologies without coordinates.
        """
        return _to_meters(self.z_)

    @property
    def x_(self):
        """
        The x coordinate (as a unitless floating point number) at the midpoint
        of each compartment in this sub-section. Returns ``None`` for
        morphologies without coordinates.
        """
        return _from_morphology(self._morphology.x_, self._i, self._j)

    @property
    def y_(self):
        """
        The y coordinate (as a unitless floating point number) at the midpoint
        of each compartment in this sub-section. Returns ``None`` for
        morphologies without coordinates.
        """
        return _from_morphology(self._morphology.y_, self._i, self._j)

    @property
    def z_(self):
        """
        The z coordinate (as a unitless floating point number) at the midpoint
        of each compartment in this sub-section. Returns ``None`` for
        morphologies without coordinates.
        """
        return _from_morphology(self._morphology.z_, self._i, self._j)

    @property
    def end_x(self):
        """
        The x coordinate at the end of each compartment in this sub-section.
        Returns ``None`` for morphologies without coordinates.
        """
        return _to_meters(self.end_x_)

    @property
    def end_y(self):
        """
        The y coordinate at the end of each compartment in this sub-section.
        Returns ``None`` for morphologies without coordinates.
        """
        return _to_meters(self.end_y_)

    @property
    def end_z(self):
        """
        The z coordinate at the end of each compartment in this sub-section.
        Returns ``None`` for morphologies without coordinates.
        """
        return _to_meters(self.end_z_)

    @property
    def end_x_(self):
        """
        The x coordinate (as a unitless floating point number) at the end of
        each compartment in this sub-section. Returns ``None`` for morphologies
        without coordinates.
        """
        return _from_morphology(self._morphology.end_x_, self._i, self._j)

    @property
    def end_y_(self):
        """
        The y coordinate (as a unitless floating point number) at the end of
        each compartment in this sub-section. Returns ``None`` for morphologies
        without coordinates.
        """
        return _from_morphology(self._morphology.end_y_, self._i, self._j)

    @property
    def end_z_(self):
        """
        The z coordinate (as a unitless floating point number) at the end of
        each compartment in this sub-section. Returns ``None`` for morphologies
        without coordinates.
        """
        return _from_morphology(self._morphology.end_z_, self._i, self._j)


class Soma(Morphology):
    """
    A spherical, iso-potential soma.

    Parameters
    ----------
    diameter : `Quantity`
        Diameter of the sphere.
    x : `Quantity`, optional
        The x coordinate of the position of the soma.
    y : `Quantity`, optional
        The y coordinate of the position of the soma.
    z : `Quantity`, optional
        The z coordinate of the position of the soma.
    type : str, optional
        The ``type`` of this section, defaults to ``'soma'``.
    """

    @check_units(diameter=meter, x=meter, y=meter, z=meter)
    def __init__(self, diameter, x=None, y=None, z=None, type="soma"):
        Morphology.__init__(self, n=1, type=type)
        if diameter.shape != () and len(diameter) != 1:
            raise TypeError("Diameter has to be a scalar value.")
        for coord in [x, y, z]:
            if coord is not None and coord.shape != () and len(coord) != 1:
                raise TypeError("Coordinates have to be scalar values.")
        self._diameter = np.ones(1) * diameter
        if any(coord is not None for coord in (x, y, z)):
            default_value = np.array([0.0])
        else:
            default_value = None
        self._x = np.atleast_1d(np.asarray(x)) if x is not None else default_value
        self._y = np.atleast_1d(np.asarray(y)) if y is not None else default_value
        self._z = np.atleast_1d(np.asarray(z)) if z is not None else default_value

    def __repr__(self):
        s = f"{self.__class__.__name__}(diameter={self.diameter[0]!r}"
        if self._x is not None:
            s += f", x={self.x[0]!r}, y={self.y[0]!r}, z={self.z[0]!r}"
        if self.type != "soma":
            s += f", type={self.type!r}"
        return f"{s})"

    def copy_section(self):
        return Soma(self.diameter, x=self.x, y=self.y, z=self.z, type=self.type)

    # Note that the per-compartment properties should always return 1D arrays,
    # i.e. for the soma arrays of length 1 instead of scalar values
    @property
    def area(self):
        """
        The membrane surface area of this section (as an array of length 1).
        """
        return np.pi * self.diameter**2

    @property
    def volume(self):
        """
        The volume of this section (as an array of length 1).
        """
        return (np.pi * self.diameter**3) / 6

    @property
    def length(self):
        """
        The "length" (equal to `diameter`) of this section (as an array of
        length 1).
        """
        return self.diameter

    @property
    def r_length_1(self):
        """
        The geometry-dependent term to calculate the conductance between the
        start and the midpoint of each compartment. Returns a fixed (high)
        value for a `Soma`, corresponding to a section with very low
        intracellular resistance.
        """
        return [1] * meter

    @property
    def r_length_2(self):
        """
        The geometry-dependent term to calculate the conductance between the
        midpoint and the end of each compartment. Returns a fixed (high)
        value for a `Soma`, corresponding to a section with very low
        intracellular resistance.
        """
        return [1] * meter

    @property
    def diameter(self):
        """
        The diameter of this section (as an array of length 1).
        """
        return self._diameter

    @property
    def distance(self):
        """
        The total distance between the midpoint of this section and the root
        of the morphology. The `Soma` is most likely the root of the
        morphology, and therefore the `distance` is 0.
        """
        dist = self._parent.distance[-1:] if self._parent is not None else [0] * um
        return dist

    @property
    def start_x_(self):
        """
        The x-coordinate of the current section (as an array of length 1). Note
        that a `Soma` is modelled as a "point" with finite surface/volume,
        equivalent to that of a sphere with the given `diameter`. It's start-,
        midpoint-, and end-coordinates are therefore identical.
        """
        return self._x

    @property
    def start_y_(self):
        """
        The y-coordinate of the current section (as an array of length 1). Note
        that a `Soma` is modelled as a "point" with finite surface/volume,
        equivalent to that of a sphere with the given `diameter`. It's start-,
        midpoint-, and end-coordinates are therefore identical.
        """
        return self._y

    @property
    def start_z_(self):
        """
        The z-coordinate of the current section (as an array of length 1). Note
        that a `Soma` is modelled as a "point" with finite surface/volume,
        equivalent to that of a sphere with the given `diameter`. It's start-,
        midpoint-, and end-coordinates are therefore identical.
        """
        return self._z

    @property
    def x_(self):
        """
        The x-coordinate of the current section (as an array of length 1). Note
        that a `Soma` is modelled as a "point" with finite surface/volume,
        equivalent to that of a sphere with the given `diameter`. It's start-,
        midpoint-, and end-coordinates are therefore identical.
        """
        return self._x

    @property
    def y_(self):
        """
        The y-coordinate of the current section (as an array of length 1). Note
        that a `Soma` is modelled as a "point" with finite surface/volume,
        equivalent to that of a sphere with the given `diameter`. It's start-,
        midpoint-, and end-coordinates are therefore identical.
        """
        return self._y

    @property
    def z_(self):
        """
        The z-coordinate of the current section (as an array of length 1). Note
        that a `Soma` is modelled as a "point" with finite surface/volume,
        equivalent to that of a sphere with the given `diameter`. It's start-,
        midpoint-, and end-coordinates are therefore identical.
        """
        return self._z

    @property
    def end_x_(self):
        """
        The x-coordinate of the current section (as an array of length 1). Note
        that a `Soma` is modelled as a "point" with finite surface/volume,
        equivalent to that of a sphere with the given `diameter`. It's start-,
        midpoint-, and end-coordinates are therefore identical.
        """
        return self._x

    @property
    def end_y_(self):
        """
        The y-coordinate of the current section (as an array of length 1). Note
        that a `Soma` is modelled as a "point" with finite surface/volume,
        equivalent to that of a sphere with the given `diameter`. It's start-,
        midpoint-, and end-coordinates are therefore identical.
        """
        return self._y

    @property
    def end_z_(self):
        """
        The z-coordinate of the current section (as an array of length 1). Note
        that a `Soma` is modelled as a "point" with finite surface/volume,
        equivalent to that of a sphere with the given `diameter`. It's start-,
        midpoint-, and end-coordinates are therefore identical.
        """
        return self._z

    @property
    def end_distance(self):
        """
        The distance to the root of the morphology at the end of this section.
        Note that since a `Soma` is modeled as a point (see docs of `x`, etc.),
        it does not add anything to the total distance, e.g. a section
        connecting to a `Soma` has a `distance` of 0 um at its start.
        """
        dist = self._parent.end_distance if self._parent is not None else 0 * um
        return dist


class Section(Morphology):
    """
    A section (unbranched structure), described as a sequence of truncated
    cones with potentially varying diameters and lengths per compartment.

    Parameters
    ----------
    diameter : `Quantity`
        Either a single value (the constant diameter along the whole section),
        or a value of length ``n+1``. When ``n+1`` values are given, they
        will be interpreted as the diameters at the start of the first
        compartment and the diameters at the end of each compartment (which is
        equivalent to: the diameter at the start of each compartment and the
        diameter at the end of the last compartment.
    n : int, optional
        The number of compartments in this section. Defaults to 1.
    length : `Quantity`, optional
        Either a single value (the total length of the section), or a value of
        length ``n``, the length of each individual compartment. Cannot be
        combined with the specification of coordinates.
    x : `Quantity`, optional
        ``n+1`` values, specifying the x coordinates of the start point of the
        first compartment and the end-points of all compartments (which is
        equivalent to: the start point of all compartments and the end point of
        the last compartment). The coordinates are interpreted as relative to
        the end point of the parent compartment (if any), so in most cases the
        start point should be ``0*um``. The common exception is a cylinder
        connecting to a `Soma`, here the start point can be used to make the
        cylinder start at the surface of the sphere instead of at its center.
        You can specify all of ``x``, ``y``, or ``z`` to specify
        a morphology in 3D, or only one or two out of them to specify a
        morphology in 1D or 2D.
    y : `Quantity`, optional
        See ``x``
    z : `Quantity`, optional
        See ``x``
    type : str, optional
        The type (e.g. ``"axon"``) of this `Section`.
    """

    @check_units(
        n=1,
        length=meter,
        diameter=meter,
        start_diameter=meter,
        x=meter,
        y=meter,
        z=meter,
    )
    def __init__(
        self,
        diameter,
        n=1,
        length=None,
        x=None,
        y=None,
        z=None,
        start_diameter=None,
        origin=None,
        type=None,
    ):
        n = int(n)
        Morphology.__init__(self, n=n, type=type)

        if diameter.ndim != 1 or len(diameter) != n + 1:
            raise TypeError(
                "The diameter argument has to be a one-dimensional array of length "
                f"{int(n + 1)}"
            )
        self._diameter = Quantity(diameter, copy=True).reshape((n + 1,))

        if (x is not None or y is not None or z is not None) and length is not None:
            raise TypeError("Cannot specify coordinates and length at the same time.")

        if length is not None:
            # Length
            if length.ndim != 1 or len(length) != n:
                raise TypeError(
                    "The length argument has to be a one-dimensional array of length "
                    f"{int(n)}"
                )
            self._length = Quantity(length, copy=True).reshape((n,))
            self._x = self._y = self._z = None
        else:
            # Coordinates
            if x is None and y is None and z is None:
                raise TypeError(
                    "No length specified, need to specify at least "
                    "one out of x, y, or z."
                )
            for name, value in [("x", x), ("y", y), ("z", z)]:
                if value is not None and (value.ndim != 1 or len(value) != n + 1):
                    raise TypeError(
                        f"'{name}' needs to be a 1-dimensional array of length {n + 1}."
                    )
            self._x = (
                np.asarray(x).reshape((n + 1,)) if x is not None else np.zeros(n + 1)
            )
            self._y = (
                np.asarray(y).reshape((n + 1,)) if y is not None else np.zeros(n + 1)
            )
            self._z = (
                np.asarray(z).reshape((n + 1,)) if z is not None else np.zeros(n + 1)
            )

            length = np.sqrt(
                (self.end_x - self.start_x) ** 2
                + (self.end_y - self.start_y) ** 2
                + (self.end_z - self.start_z) ** 2
            )
            self._length = length

    def __repr__(self):
        if all(
            np.abs(self.end_diameter - self.end_diameter[0])
            < self.end_diameter[0] * 1e-12
        ):
            # Constant diameter
            diam = self.end_diameter[0]
        else:
            diam = (
                np.hstack(
                    [np.asarray(self.start_diameter[0]), np.asarray(self.end_diameter)]
                )
                * meter
            )
        s = f"{self.__class__.__name__}(diameter={diam!r}"
        if self.n != 1:
            s += f", n={self.n}"
        if self._x is not None:
            s += f", x={self._x!r}, y={self._y!r}, z={self._z!r}"
        else:
            s += f", length={sum(self._length)!r}"
        if self.type is not None:
            s += f", type={self.type!r}"
        return f"{s})"

    def copy_section(self):
        if self.x is None:
            x, y, z = None, None, None
            length = self.length
        else:
            x, y, z = self._x * meter, self._y * meter, self._z * meter
            length = None
        return Section(
            diameter=self._diameter,
            n=self.n,
            x=x,
            y=y,
            z=z,
            length=length,
            type=self.type,
        )

    @property
    def area(self):
        r"""
        The membrane surface area of each compartment in this section. The
        surface area of each compartment is calculated as
        :math:`\frac{\pi}{2}(d_1 + d_2)\sqrt{\frac{(d_1 - d_2)^2}{4} + l^2)}`,
        where :math:`l` is the length of the compartment, and :math:`d_1` and
        :math:`d_2` are the diameter at the start and end of the compartment,
        respectively. Note that this surface area does not contain the area of
        the two disks at the two sides of the truncated cone.
        """
        d_1 = self.start_diameter
        d_2 = self.end_diameter
        return (
            np.pi / 2 * (d_1 + d_2) * np.sqrt(((d_1 - d_2) ** 2) / 4 + self._length**2)
        )

    @property
    def volume(self):
        r"""
        The volume of each compartment in this section. The volume of each
        compartment is calculated as
        :math:`\frac{\pi}{12} l (d_1^2 + d_1 d_2 + d_2^2)`,
        where :math:`l` is the length of the compartment, and :math:`d_1` and
        :math:`d_2` are the diameter at the start and end of the compartment,
        respectively.
        """
        d_1 = self.start_diameter
        d_2 = self.end_diameter
        return np.pi * self._length * (d_1**2 + d_1 * d_2 + d_2**2) / 12

    @property
    def length(self):
        """
        The length of each compartment in this section.
        """
        return self._length

    @property
    def start_diameter(self):
        """
        The diameter at the start of each compartment in this section.
        """
        return Quantity(self._diameter[:-1], copy=True)

    @property
    def end_diameter(self):
        """
        The diameter at the end of each compartment in this section.
        """
        return Quantity(self._diameter[1:], copy=True)

    @property
    def diameter(self):
        """
        The diameter at the middle of each compartment in this section.
        """
        d_1 = self.start_diameter
        d_2 = self.end_diameter
        # Diameter at the center
        return 0.5 * (d_1 + d_2)

    @property
    def distance(self):
        """
        The total distance between the midpoint of each compartment and the root
        of the morphology.
        """
        dist = self._parent.end_distance if self._parent is not None else 0 * um
        return dist + np.cumsum(self.length) - 0.5 * self.length

    @property
    def end_distance(self):
        """
        The distance to the root of the morphology at the end of this section.
        """
        return self.distance[-1] + 0.5 * self.length[-1]

    @property
    def r_length_1(self):
        """
        The geometry-dependent term to calculate the conductance between the
        start and the midpoint of each compartment. Dividing this value by the
        Intracellular resistivity gives the conductance.
        """
        d_1 = self.start_diameter
        d_2 = (self.start_diameter + self.end_diameter) * 0.5
        return np.pi / 2 * (d_1 * d_2) / self._length

    @property
    def r_length_2(self):
        """
        The geometry-dependent term to calculate the conductance between the
        midpoint and the end of each compartment. Dividing this value by the
        Intracellular resistivity gives the conductance.
        """
        d_1 = (self.start_diameter + self.end_diameter) * 0.5
        d_2 = self.end_diameter
        return np.pi / 2 * (d_1 * d_2) / self._length

    @property
    def start_x_(self):
        """
        The x coordinate (as a unitless floating point number) at the beginning
        of each compartment. Returns ``None`` for morphologies without
        coordinates.
        """
        if self._x is None:
            return None
        parent_end_x = self.parent.end_x_ if self.parent is not None else None
        if parent_end_x is not None:
            return parent_end_x[-1] + self._x[:-1]
        else:
            return self._x[:-1]

    @property
    def start_y_(self):
        """
        The y coordinate (as a unitless floating point number) at the beginning
        of each compartment. Returns ``None`` for morphologies without
        coordinates.
        """
        if self._y is None:
            return None
        parent_end_y = self.parent.end_y_ if self.parent is not None else None
        if parent_end_y is not None:
            return parent_end_y[-1] + self._y[:-1]
        else:
            return self._y[:-1]

    @property
    def start_z_(self):
        """
        The z coordinate (as a unitless floating point number) at the beginning
        of each compartment. Returns ``None`` for morphologies without
        coordinates.
        """
        if self._z is None:
            return None
        parent_end_z = self.parent.end_z_ if self.parent is not None else None
        if parent_end_z is not None:
            return parent_end_z[-1] + self._z[:-1]
        else:
            return self._z[:-1]

    @property
    def x_(self):
        """
        The x coordinate (as a unitless floating point number) at the midpoint
        of each compartment. Returns ``None`` for morphologies without
        coordinates.
        """
        if self._x is None:
            return None
        start_x = self.start_x_
        diff_x = self.end_x_ - start_x
        return start_x + 0.5 * diff_x

    @property
    def y_(self):
        """
        The y coordinate (as a unitless floating point number) at the midpoint
        of each compartment. Returns ``None`` for morphologies without
        coordinates.
        """
        if self._y is None:
            return None
        start_y = self.start_y_
        diff_y = self.end_y_ - start_y
        return start_y + 0.5 * diff_y

    @property
    def z_(self):
        """
        The z coordinate (as a unitless floating point number) at the midpoint
        of each compartment. Returns ``None`` for morphologies without
        coordinates.
        """
        if self._z is None:
            return None
        start_z = self.start_z_
        diff_z = self.end_z_ - start_z
        return start_z + 0.5 * diff_z

    @property
    def end_x_(self):
        """
        The x coordinate (as a unitless floating point number) at the end of
        each compartment. Returns ``None`` for morphologies without coordinates.
        """
        if self._x is None:
            return None
        parent_end_x = self.parent.end_x_ if self.parent is not None else None
        if parent_end_x is not None:
            return parent_end_x[-1] + self._x[1:]
        else:
            return self._x[1:]

    @property
    def end_y_(self):
        """
        The y coordinate (as a unitless floating point number) at the end of
        each compartment. Returns ``None`` for morphologies without coordinates.
        """
        if self._y is None:
            return None
        parent_end_y = self.parent.end_y_ if self.parent is not None else None
        if parent_end_y is not None:
            return parent_end_y[-1] + self._y[1:]
        else:
            return self._y[1:]

    @property
    def end_z_(self):
        """
        The z coordinate (as a unitless floating point number) at the end of
        each compartment. Returns ``None`` for morphologies without coordinates.
        """
        if self._z is None:
            return None
        parent_end_z = self.parent.end_z_ if self.parent is not None else None
        if parent_end_z is not None:
            return parent_end_z[-1] + self._z[1:]
        else:
            return self._z[1:]


class Cylinder(Section):
    """
    A cylindrical section. For sections with more complex geometry (varying
    length and/or diameter of each compartment), use the `Section` class.

    Parameters
    ----------
    diameter : `Quantity`
        The diameter of the cylinder.
    n : int, optional
        The number of compartments in this section. Defaults to 1.
    length : `Quantity`, optional
        The length of the cylinder. Cannot be combined with the specification
        of coordinates.
    x : `Quantity`, optional
        A sequence of two values, the start and the end point of the cylinder.
        The coordinates are interpreted as relative to the end point of the
        parent compartment (if any), so in most cases the start point should
        be ``0*um``. The common exception is a cylinder connecting to a `Soma`,
        here the start point can be used to make the cylinder start at the
        surface of the sphere instead of at its center.
        You can specify all of ``x``, ``y``, or ``z`` to specify
        a morphology in 3D, or only one or two out of them to specify a
        morphology in 1D or 2D.
    y : `Quantity`, optional
        See ``x``
    z : `Quantity`, optional
        See ``x``
    type : str, optional
        The type (e.g. ``"axon"``) of this `Cylinder`.
    """

    @check_units(n=1, length=meter, diameter=meter, x=meter, y=meter, z=meter)
    def __init__(self, diameter, n=1, length=None, x=None, y=None, z=None, type=None):
        n = int(n)
        Morphology.__init__(self, n=n, type=type)

        # Diameter
        if diameter.shape != () and (diameter.ndim > 1 or len(diameter) != 1):
            raise TypeError("The diameter argument has to be a single value.")
        diameter = np.ones(n) * diameter
        self._diameter = diameter

        if (x is not None or y is not None or z is not None) and length is not None:
            raise TypeError("Cannot specify coordinates and length at the same time.")

        if length is not None:
            # Length
            if length.shape != () and (length.ndim > 1 or len(length) != 1):
                raise TypeError("The length argument has to be a single value.")
            self._length = np.ones(n) * (length / n)  # length was total length
            self._x = self._y = self._z = None
        else:
            # Coordinates
            if x is None and y is None and z is None:
                raise TypeError(
                    "No length specified, need to specify at least "
                    "one out of x, y, or z."
                )
            for name, value in [("x", x), ("y", y), ("z", z)]:
                if value is not None and (value.ndim != 1 or len(value) != 2):
                    raise TypeError(
                        f"{name} needs to be a 1-dimensional array of length 2 (start "
                        "and end point)"
                    )
            self._x = (
                np.asarray(np.linspace(x[0], x[1], n + 1))
                if x is not None
                else np.zeros(n + 1)
            )
            self._y = (
                np.asarray(np.linspace(y[0], y[1], n + 1))
                if y is not None
                else np.zeros(n + 1)
            )
            self._z = (
                np.asarray(np.linspace(z[0], z[1], n + 1))
                if z is not None
                else np.zeros(n + 1)
            )
            length = np.sqrt(
                (self.end_x - self.start_x) ** 2
                + (self.end_y - self.start_y) ** 2
                + (self.end_z - self.start_z) ** 2
            )
            self._length = length

    def __repr__(self):
        s = f"{self.__class__.__name__}(diameter={self.diameter[0]!r}"
        if self.n != 1:
            s += f", n={self.n}"
        if self._x is not None:
            s += (
                f", x={self._x[[0, -1]]!r}, y={self._y[[0, -1]]!r},"
                f" z={self._z[[0, -1]]!r}"
            )
        else:
            s += f", length={sum(self._length)!r}"
        if self.type is not None:
            s += f", type={self.type!r}"
        return f"{s})"

    def copy_section(self):
        if self.x is None:
            return Cylinder(
                self.diameter[0], n=self.n, length=self.length, type=self.type
            )
        else:
            return Cylinder(
                self.diameter[0],
                n=self.n,
                x=self._x[[0, -1]],
                y=self._y[[0, -1]],
                z=self._z[[0, -1]],
                type=self.type,
            )

    # Overwrite the properties that differ from `Section`
    @property
    def area(self):
        r"""
        The membrane surface area of each compartment in this section. The
        surface area of each compartment is calculated as
        :math:`\pi d l`,
        where :math:`l` is the length of the compartment, and :math:`d` is its
        diameter. Note that this surface area does not contain the area of
        the two disks at the two sides of the cylinder.
        """
        return np.pi * self._diameter * self.length

    @property
    def start_diameter(self):
        """
        The diameter at the start of each compartment in this section.
        """
        return self._diameter

    @property
    def diameter(self):
        """
        The diameter at the middle of each compartment in this section.
        """
        return self._diameter

    @property
    def end_diameter(self):
        """
        The diameter at the end of each compartment in this section.
        """
        return self._diameter

    @property
    def volume(self):
        r"""
        The volume of each compartment in this section. The volume of each
        compartment is calculated as
        :math:`\pi \frac{d}{2}^2 l` ,
        where :math:`l` is the length of the compartment, and :math:`d` is its
        diameter.
        """
        return np.pi * (self._diameter / 2) ** 2 * self.length

    @property
    def r_length_1(self):
        """
        The geometry-dependent term to calculate the conductance between the
        start and the midpoint of each compartment. Dividing this value by the
        Intracellular resistivity gives the conductance.
        """
        return np.pi / 2 * (self._diameter**2) / self.length

    @property
    def r_length_2(self):
        """
        The geometry-dependent term to calculate the conductance between the
        midpoint and the end of each compartment. Dividing this value by the
        Intracellular resistivity gives the conductance.
        """
        return np.pi / 2 * (self._diameter**2) / self.length



import time
from processify import processify

def halfway_timer():
    global halftime
    halftime = time.time()

@processify
def brian1_CUBA(N, duration=1,
                do_refractory=False, exact_method=False, do_monitor=False, do_synapses=False,
                **ignored_opts):
    global halftime
    import time
    results = dict()
    start_time = time.time()
    from brian import *
    finished_importing_brian_time = time.time()
    results['import_brian'] = finished_importing_brian_time-start_time

    duration *= second
    Ne = int(0.8*N)
    Ni = N-Ne
    taum = 20 * ms
    taue = 5 * ms
    taui = 10 * ms
    Vt = -50 * mV
    Vr = -60 * mV
    El = -49 * mV

    eqs = '''
    dv/dt  = (ge+gi-(v-El))/taum : volt
    dge/dt = -ge/taue : volt
    dgi/dt = -gi/taui : volt
    '''

    kwds = {}
    if do_refractory:
        kwds['refractory'] = 5*ms
    if not exact_method:
        kwds['method'] = 'Euler'
    P = NeuronGroup(N, model=eqs, threshold=Vt, reset=Vr, **kwds)
    P.v = Vr
    P.ge = 0 * mV
    P.gi = 0 * mV

    Pe = P.subgroup(Ne)
    Pi = P.subgroup(Ni)
    we = (60 * 0.27 / 10) * mV # excitatory synaptic weight (voltage)
    wi = (-20 * 4.5 / 10) * mV # inhibitory synaptic weight
    p = min(80./N, 1)
    if do_synapses:
        Ce = Connection(Pe, P, 'ge', weight=we, sparseness=p)
        Ci = Connection(Pi, P, 'gi', weight=wi, sparseness=p)
    P.v = Vr + rand(len(P)) * (Vt - Vr)

    if do_monitor:
        M = SpikeMonitor(P)

    netop_halfway_timer = network_operation(clock=EventClock(dt=duration*0.51))(halfway_timer)

    objects_created_time = time.time()
    results['object_creation'] = objects_created_time-finished_importing_brian_time

    run(1 * msecond)

    initial_run_time = time.time()
    results['initial_run'] = initial_run_time-objects_created_time


    run(duration)

    main_run_time = time.time()
    results['main_run'] = main_run_time-initial_run_time
    results['second_half_main_run'] = main_run_time-halftime
    results['total'] = main_run_time-start_time

    return results

@processify
def brian2_CUBA(N, duration=1,
                do_refractory=False, exact_method=False, do_monitor=False, do_synapses=False,
                codegen_target='numpy',
                **ignored_opts):
    global halftime
    import time
    results = dict()
    start_time = time.time()
    from brian2 import *
    prefs.codegen.target = codegen_target
    finished_importing_brian_time = time.time()
    results['import_brian'] = finished_importing_brian_time-start_time

    duration *= second
    Ne = int(0.8*N)
    Ni = N-Ne
    taum = 20 * ms
    taue = 5 * ms
    taui = 10 * ms
    Vt = -50 * mV
    Vr = -60 * mV
    El = -49 * mV

    eqs = '''
    dv/dt  = (ge+gi-(v-El))/taum : volt (unless refractory)
    dge/dt = -ge/taue : volt
    dgi/dt = -gi/taui : volt
    '''
    if not do_refractory:
        eqs = eqs.replace('(unless refractory)', '')

    kwds = {}
    if do_refractory:
        kwds['refractory'] = 5*ms
    if not exact_method:
        kwds['method'] = 'euler'
    P = NeuronGroup(N, model=eqs, threshold='v>Vt', reset='v=Vr', **kwds)
    P.v = Vr
    P.ge = 0 * mV
    P.gi = 0 * mV

    we = (60 * 0.27 / 10) * mV # excitatory synaptic weight (voltage)
    wi = (-20 * 4.5 / 10) * mV # inhibitory synaptic weight
    p = min(80./N, 1.0)
    if do_synapses:
        Ce = Synapses(P, P, on_pre='ge += we')
        Ci = Synapses(P, P, on_pre='gi += wi')
        Ce.connect('i<Ne', p=p)
        Ci.connect('i>=Ne', p=p)
    P.v = Vr + rand(len(P)) * (Vt - Vr)

    if do_monitor:
        M = SpikeMonitor(P)

    netop_halfway_timer = NetworkOperation(halfway_timer, dt=duration*0.51)

    objects_created_time = time.time()
    results['object_creation'] = objects_created_time-finished_importing_brian_time

    run(1 * msecond)

    initial_run_time = time.time()
    results['initial_run'] = initial_run_time-objects_created_time

    run(duration)

    main_run_time = time.time()
    results['main_run'] = main_run_time-initial_run_time
    results['second_half_main_run'] = main_run_time-halftime
    results['total'] = main_run_time-start_time

    return results

def brian2_CUBA_cython(*args, **opts):
    opts['codegen_target'] = 'cython'
    return brian2_CUBA(*args, **opts)

if __name__=='__main__':
    import functools
    from pylab import *
    numfigs = 0

    funcs = [
        brian1_CUBA,
        brian2_CUBA,
        brian2_CUBA_cython,
        ]
    options = dict(duration=1,
                   do_monitor=False,
                   do_refractory=False,
                   do_synapses=False,
                   exact_method=False,
                   )
    N = [1, 10, 100, 1000,
         10000,
         #100000,
         ]

    for name in options.keys()+['']:
    #for name in ['']:
        if name:
            if not isinstance(options[name], bool):
                continue
            options[name] = True
        figure(figsize=(16, 8))
        for func in funcs:
            pfunc = functools.partial(func, **options)
            all_results = map(pfunc, N)
            times = collections.defaultdict(list)
            for res in all_results:
                for k, v in res.items():
                    times[k].append(v)
            for i, k in enumerate(sorted(times.keys())):
                subplot(2, 3, i+1)
                if i>numfigs:
                    numfigs = i
                title(k)
                v = times[k]
                loglog(N[1:], v[1:], label=func.__name__)
        for i in range(numfigs+1):
            subplot(2, 3, i+1)
            legend(loc='best')
        suptitle(', '.join('%s=%s' % (k, options[k]) for k in sorted(options.keys())))
        tight_layout()
        subplots_adjust(top=0.9)
        if name:
            savefig('compare_to_brian1.%s.png'%name)
            options[name] = False
        else:
            savefig('compare_to_brian1.png')
    show()



import pytest
from numpy.testing import assert_equal

from brian2 import *
from brian2.core.network import schedule_propagation_offset
from brian2.devices.device import reinit_and_delete
from brian2.tests.utils import assert_allclose, exc_isinstance


@pytest.mark.standalone_compatible
def test_poissoninput():
    # Test extreme cases and do a very basic test of an intermediate case, we
    # don't want tests to be stochastic
    G = NeuronGroup(
        10,
        """
        x : volt
        y : volt
        y2 : volt
        z : volt
        z2 : volt
        w : 1
        """,
    )
    G.w = 0.5

    never_update = PoissonInput(G, "x", 100, 0 * Hz, weight=1 * volt)
    always_update = PoissonInput(G, "y", 50, 1 / defaultclock.dt, weight=2 * volt)
    always_update2 = PoissonInput(
        G, "y2", 50, 1 / defaultclock.dt, weight="1*volt + 1*volt"
    )
    sometimes_update = PoissonInput(G, "z", 10000, 50 * Hz, weight=0.5 * volt)
    sometimes_update2 = PoissonInput(G, "z2", 10000, 50 * Hz, weight="w*volt")

    assert_equal(never_update.rate, 0 * Hz)
    assert_equal(never_update.N, 100)
    assert_equal(always_update.rate, 1 / defaultclock.dt)
    assert_equal(always_update.N, 50)
    assert_equal(sometimes_update.rate, 50 * Hz)
    assert_equal(sometimes_update.N, 10000)

    mon = StateMonitor(G, ["x", "y", "y2", "z", "z2"], record=True, when="end")

    run(1 * ms)
    assert_equal(0, mon.x[:])
    assert_equal(
        np.tile((1 + np.arange(mon.y[:].shape[1])) * 50 * 2 * volt, (10, 1)), mon.y[:]
    )
    assert_equal(
        np.tile((1 + np.arange(mon.y[:].shape[1])) * 50 * 2 * volt, (10, 1)), mon.y2[:]
    )
    assert all(np.var(np.diff(mon.z[:]), axis=1) > 0)  # variability over time
    assert all(np.var(mon.z[:], axis=0) > 0)  # variability over neurons
    assert all(np.var(np.diff(mon.z2[:]), axis=1) > 0)  # variability over time
    assert all(np.var(mon.z2[:], axis=0) > 0)  # variability over neurons


@pytest.mark.codegen_independent
def test_poissoninput_errors():
    # Targeting non-existing variable
    G = NeuronGroup(
        10,
        """
        x : volt
        y : 1
        """,
    )
    with pytest.raises(KeyError):
        PoissonInput(G, "z", 100, 100 * Hz, weight=1.0)

    # Incorrect units
    with pytest.raises(DimensionMismatchError):
        PoissonInput(G, "x", 100, 100 * Hz, weight=1.0)
    with pytest.raises(DimensionMismatchError):
        PoissonInput(G, "y", 100, 100 * Hz, weight=1.0 * volt)

    # dt change
    old_dt = defaultclock.dt
    inp = PoissonInput(G, "x", 100, 100 * Hz, weight=1 * volt)
    defaultclock.dt = 2 * old_dt
    net = Network(collect())
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms)
    assert exc_isinstance(exc, NotImplementedError)
    defaultclock.dt = old_dt


@pytest.mark.standalone_compatible
def test_poissoninput_refractory():
    eqs = """
    dv/dt = 0/second : 1 (unless refractory)
    """
    G = NeuronGroup(
        10, eqs, reset="v=0", threshold="v>4.5", refractory=5 * defaultclock.dt
    )
    # Will increase the value by 1.0 at each time step
    P = PoissonInput(G, "v", 1, 1 / defaultclock.dt, weight=1.0)
    mon = StateMonitor(G, "v", record=5)
    run(10 * defaultclock.dt)
    expected = np.arange(10, dtype=float)
    expected[6 - int(schedule_propagation_offset() / defaultclock.dt) :] = 0
    assert_allclose(mon[5].v[:], expected)


if __name__ == "__main__":
    test_poissoninput()
    reinit_and_delete()
    test_poissoninput_errors()
    test_poissoninput_refractory()



import sympy as sp

from brian2.parsing.sympytools import str_to_sympy, sympy_to_str

from .base import (
    StateUpdateMethod,
    UnsupportedEquationsException,
    extract_method_options,
)

__all__ = ["exponential_euler"]


def get_conditionally_linear_system(eqs, variables=None):
    """
    Convert equations into a linear system using sympy.

    Parameters
    ----------
    eqs : `Equations`
        The model equations.

    Returns
    -------
    coefficients : dict of (sympy expression, sympy expression) tuples
        For every variable x, a tuple (M, B) containing the coefficients M and
        B (as sympy expressions) for M * x + B

    Raises
    ------
    ValueError
        If one of the equations cannot be converted into a M * x + B form.

    Examples
    --------
    >>> from brian2 import Equations
    >>> eqs = Equations('''
    ... dv/dt = (-v + w**2.0) / tau : 1
    ... dw/dt = -w / tau : 1
    ... ''')
    >>> system = get_conditionally_linear_system(eqs)
    >>> print(system['v'])
    (-1/tau, w**2.0/tau)
    >>> print(system['w'])
    (-1/tau, 0)

    """
    diff_eqs = eqs.get_substituted_expressions(variables)

    coefficients = {}

    for name, expr in diff_eqs:
        var = sp.Symbol(name, real=True)

        s_expr = str_to_sympy(expr.code, variables).expand()
        if s_expr.has(var):
            # Factor out the variable
            s_expr = sp.collect(s_expr, var, evaluate=False)

            if len(s_expr) > 2 or var not in s_expr:
                raise ValueError(
                    f"The expression '{expr}', defining the variable "
                    f"'{name}', could not be separated into linear "
                    "components."
                )
            coefficients[name] = (s_expr[var], s_expr.get(1, 0))
        else:
            coefficients[name] = (0, s_expr)

    return coefficients


class ExponentialEulerStateUpdater(StateUpdateMethod):
    """
    A state updater for conditionally linear equations, i.e. equations where
    each variable only depends linearly on itself (but possibly non-linearly
    on other variables). Typical Hodgkin-Huxley equations fall into this
    category, it is therefore the default integration method used in the
    GENESIS simulator, for example.
    """

    def __call__(self, equations, variables=None, method_options=None):
        extract_method_options(method_options, {})
        if equations.is_stochastic:
            raise UnsupportedEquationsException(
                "Cannot solve stochastic equations with this state updater."
            )

        # Try whether the equations are conditionally linear
        try:
            system = get_conditionally_linear_system(equations, variables)
        except ValueError:
            raise UnsupportedEquationsException(
                "Can only solve conditionally linear systems with this state updater."
            )

        code = []
        for var, (A, B) in system.items():
            s_var = sp.Symbol(var)
            s_dt = sp.Symbol("dt")
            if A == 0:
                update_expression = s_var + s_dt * B
            elif B != 0:
                BA = B / A
                # Avoid calculating B/A twice
                BA_name = f"_BA_{var}"
                s_BA = sp.Symbol(BA_name)
                code += [f"{BA_name} = {sympy_to_str(BA)}"]
                update_expression = (s_var + s_BA) * sp.exp(A * s_dt) - s_BA
            else:
                update_expression = s_var * sp.exp(A * s_dt)

            # The actual update step
            code += [f"_{var} = {sympy_to_str(update_expression)}"]

        # Replace all the variables with their updated value
        for var in system:
            code += [f"{var} = _{var}"]

        return "\n".join(code)

    # Copy doc from parent class
    __call__.__doc__ = StateUpdateMethod.__call__.__doc__


exponential_euler = ExponentialEulerStateUpdater()



from pyparsing import (
    CharsNotIn,
    Combine,
    Optional,
    ParseException,
    Regex,
    Suppress,
    Word,
    alphas,
    nums,
)

from brian2.utils.caching import cached

VARIABLE = Word(f"{alphas}_", f"{alphas + nums}_").setResultsName("variable")

OP = Regex(r"(\+|\-|\*|/|//|%|\*\*|>>|<<|&|\^|\|)?=").setResultsName("operation")
EXPR = Combine(
    CharsNotIn("=", min=1, max=1) + Optional(CharsNotIn("#"))
).setResultsName("expression")
COMMENT = Optional(CharsNotIn("#")).setResultsName("comment")
STATEMENT = VARIABLE + OP + EXPR + Optional(Suppress("#") + COMMENT)


@cached
def parse_statement(code):
    """
    parse_statement(code)

    Parses a single line of code into "var op expr".

    Parameters
    ----------
    code : str
        A string containing a single statement of the form
        ``var op expr # comment``, where the ``# comment`` part is optional.

    Returns
    -------
    var, op, expr, comment : str, str, str, str
        The four parts of the statement.

    Examples
    --------
    >>> parse_statement('v = -65*mV  # reset the membrane potential')
    ('v', '=', '-65*mV', 'reset the membrane potential')
    >>> parse_statement('v += dt*(-v/tau)')
    ('v', '+=', 'dt*(-v/tau)', '')
    """
    try:
        parsed = STATEMENT.parseString(code, parseAll=True)
    except ParseException as p_exc:
        raise ValueError(
            "Parsing the statement failed: \n"
            + str(p_exc.line)
            + "\n"
            + " " * (p_exc.column - 1)
            + "^\n"
            + str(p_exc)
        )
    parsed_statement = (
        parsed["variable"].strip(),
        parsed["operation"],
        parsed["expression"].strip(),
        parsed.get("comment", "").strip(),
    )

    return parsed_statement



import itertools
import pickle
import warnings

import numpy as np
import pytest
from numpy.testing import assert_equal

import brian2
from brian2.core.preferences import prefs
from brian2.tests.utils import assert_allclose
from brian2.units.allunits import *
from brian2.units.fundamentalunits import (
    DIMENSIONLESS,
    UFUNCS_DIMENSIONLESS,
    UFUNCS_DIMENSIONLESS_TWOARGS,
    UFUNCS_INTEGERS,
    UFUNCS_LOGICAL,
    DimensionMismatchError,
    Quantity,
    Unit,
    check_units,
    fail_for_dimension_mismatch,
    get_dimensions,
    get_or_create_dimension,
    get_unit,
    have_same_dimensions,
    in_unit,
    is_dimensionless,
    is_scalar_type,
)
from brian2.units.stdunits import Hz, cm, kHz, mM, ms, mV, nA, nS

# To work around an issue in matplotlib 1.3.1 (see
# https://github.com/matplotlib/matplotlib/pull/2591), we make `ravel`
# return a unitless array and emit a warning explaining the issue.
use_matplotlib_units_fix = False
try:
    import matplotlib

    if matplotlib.__version__ == "1.3.1":
        use_matplotlib_units_fix = True
except ImportError:
    pass


def assert_quantity(q, values, unit):
    assert isinstance(q, Quantity) or (
        have_same_dimensions(unit, 1)
        and (values.shape == () or isinstance(q, np.ndarray))
    ), q
    assert_allclose(np.asarray(q), values)
    assert have_same_dimensions(
        q, unit
    ), f"Dimension mismatch: ({get_dimensions(q)}) ({get_dimensions(unit)})"


@pytest.mark.codegen_independent
def test_construction():
    """Test the construction of quantity objects"""
    q = 500 * ms
    assert_quantity(q, 0.5, second)
    q = np.float64(500) * ms
    assert_quantity(q, 0.5, second)
    q = np.array(500) * ms
    assert_quantity(q, 0.5, second)
    q = np.array([500, 1000]) * ms
    assert_quantity(q, np.array([0.5, 1]), second)
    q = Quantity(500)
    assert_quantity(q, 500, 1)
    q = Quantity(500, dim=second.dim)
    assert_quantity(q, 500, second)
    q = Quantity([0.5, 1], dim=second.dim)
    assert_quantity(q, np.array([0.5, 1]), second)
    q = Quantity(np.array([0.5, 1]), dim=second.dim)
    assert_quantity(q, np.array([0.5, 1]), second)
    q = Quantity([500 * ms, 1 * second])
    assert_quantity(q, np.array([0.5, 1]), second)
    q = Quantity.with_dimensions(np.array([0.5, 1]), second=1)
    assert_quantity(q, np.array([0.5, 1]), second)
    q = [0.5, 1] * second
    assert_quantity(q, np.array([0.5, 1]), second)

    # dimensionless quantities
    q = Quantity([1, 2, 3])
    assert_quantity(q, np.array([1, 2, 3]), Unit(1))
    q = Quantity(np.array([1, 2, 3]))
    assert_quantity(q, np.array([1, 2, 3]), Unit(1))
    q = Quantity([])
    assert_quantity(q, np.array([]), Unit(1))

    # copying/referencing a quantity
    q1 = Quantity.with_dimensions(np.array([0.5, 1]), second=1)
    q2 = Quantity(q1)  # no copy
    assert_quantity(q2, np.asarray(q1), q1)
    q2[0] = 3 * second
    assert_equal(q1[0], 3 * second)

    q1 = Quantity.with_dimensions(np.array([0.5, 1]), second=1)
    q2 = Quantity(q1, copy=True)  # copy
    assert_quantity(q2, np.asarray(q1), q1)
    q2[0] = 3 * second
    assert_equal(q1[0], 0.5 * second)

    # Illegal constructor calls
    with pytest.raises(TypeError):
        Quantity([500 * ms, 1])
    with pytest.raises(TypeError):
        Quantity(["some", "nonsense"])
    with pytest.raises(DimensionMismatchError):
        Quantity([500 * ms, 1 * volt])


@pytest.mark.codegen_independent
def test_get_dimensions():
    """
    Test various ways of getting/comparing the dimensions of a quantity.
    """
    q = 500 * ms
    assert get_dimensions(q) is get_or_create_dimension(q.dimensions._dims)
    assert get_dimensions(q) is q.dimensions
    assert q.has_same_dimensions(3 * second)
    dims = q.dimensions
    assert_equal(dims.get_dimension("time"), 1.0)
    assert_equal(dims.get_dimension("length"), 0)

    assert get_dimensions(5) is DIMENSIONLESS
    assert get_dimensions(5.0) is DIMENSIONLESS
    assert get_dimensions(np.array(5, dtype=np.int32)) is DIMENSIONLESS
    assert get_dimensions(np.array(5.0)) is DIMENSIONLESS
    assert get_dimensions(np.float32(5.0)) is DIMENSIONLESS
    assert get_dimensions(np.float64(5.0)) is DIMENSIONLESS
    assert is_scalar_type(5)
    assert is_scalar_type(5.0)
    assert is_scalar_type(np.array(5, dtype=np.int32))
    assert is_scalar_type(np.array(5.0))
    assert is_scalar_type(np.float32(5.0))
    assert is_scalar_type(np.float64(5.0))
    with pytest.raises(TypeError):
        get_dimensions("a string")
    # wrong number of indices
    with pytest.raises(TypeError):
        get_or_create_dimension([1, 2, 3, 4, 5, 6])
    # not a sequence
    with pytest.raises(TypeError):
        get_or_create_dimension(42)


@pytest.mark.codegen_independent
def test_display():
    """
    Test displaying a quantity in different units
    """
    assert_equal(in_unit(3 * volt, mvolt), "3000. mV")
    assert_equal(in_unit(10 * mV, ohm * amp), "0.01 ohm A")
    with pytest.raises(DimensionMismatchError):
        in_unit(10 * nS, ohm)

    # A bit artificial...
    assert_equal(in_unit(10.0, Unit(10.0, scale=1)), "1.0")


@pytest.mark.codegen_independent
def test_scale():
    # Check that unit scaling is implemented correctly
    from brian2.core.namespace import DEFAULT_UNITS

    siprefixes = {
        "y": 1e-24,
        "z": 1e-21,
        "a": 1e-18,
        "f": 1e-15,
        "p": 1e-12,
        "n": 1e-9,
        "u": 1e-6,
        "m": 1e-3,
        "": 1.0,
        "k": 1e3,
        "M": 1e6,
        "G": 1e9,
        "T": 1e12,
        "P": 1e15,
        "E": 1e18,
        "Z": 1e21,
        "Y": 1e24,
    }
    for prefix in siprefixes:
        if prefix in ["c", "d", "da", "h"]:
            continue
        scaled_unit = DEFAULT_UNITS[f"{prefix}meter"]
        assert_allclose(float(scaled_unit), siprefixes[prefix])
        assert_allclose(5 * scaled_unit / meter, 5 * siprefixes[prefix])
        scaled_unit = DEFAULT_UNITS[f"{prefix}meter2"]
        assert_allclose(float(scaled_unit), siprefixes[prefix] ** 2)
        assert_allclose(5 * scaled_unit / meter2, 5 * siprefixes[prefix] ** 2)
        scaled_unit = DEFAULT_UNITS[f"{prefix}meter3"]
        assert_allclose(float(scaled_unit), siprefixes[prefix] ** 3)
        assert_allclose(5 * scaled_unit / meter3, 5 * siprefixes[prefix] ** 3)
        # liter, gram, and molar are special, they are not base units with a
        # value of one, even though they do not have any prefix
        for unit, factor in [
            ("liter", 1e-3),
            ("litre", 1e-3),
            ("gram", 1e-3),
            ("gramme", 1e-3),
            ("molar", 1e3),
        ]:
            base_unit = DEFAULT_UNITS[unit]
            scaled_unit = DEFAULT_UNITS[prefix + unit]
            assert_allclose(float(scaled_unit), siprefixes[prefix] * factor)
            assert_allclose(5 * scaled_unit / base_unit, 5 * siprefixes[prefix])


@pytest.mark.codegen_independent
def test_pickling():
    """
    Test pickling of units.
    """
    for q in [
        500 * mV,
        500 * mV / mV,
        np.arange(10) * mV,
        np.arange(12).reshape(4, 3) * mV / ms,
    ]:
        pickled = pickle.dumps(q)
        unpickled = pickle.loads(pickled)
        assert isinstance(unpickled, type(q))
        assert have_same_dimensions(unpickled, q)
        assert_equal(unpickled, q)


@pytest.mark.codegen_independent
def test_dimension_singletons():
    # Make sure that Dimension objects are singletons, even when pickled
    volt_dim = get_or_create_dimension((2, 1, -3, -1, 0, 0, 0))
    assert volt.dim is volt_dim
    import pickle

    pickled_dim = pickle.dumps(volt_dim)
    unpickled_dim = pickle.loads(pickled_dim)
    assert unpickled_dim is volt_dim
    assert unpickled_dim is volt.dim


@pytest.mark.codegen_independent
def test_str_repr():
    """
    Test that str representations do not raise any errors and that repr
    fullfills eval(repr(x)) == x. Also test generating LaTeX representations via sympy.
    """
    import sympy
    from numpy import array  # necessary for evaluating repr

    units_which_should_exist = [
        metre,
        meter,
        kilogram,
        kilogramme,
        second,
        amp,
        kelvin,
        mole,
        candle,
        radian,
        steradian,
        hertz,
        newton,
        pascal,
        joule,
        watt,
        coulomb,
        volt,
        farad,
        ohm,
        siemens,
        weber,
        tesla,
        henry,
        lumen,
        lux,
        becquerel,
        gray,
        sievert,
        katal,
        gram,
        gramme,
        molar,
        liter,
        litre,
    ]

    # scaled versions of all these units should exist (we just check farad as an example)
    some_scaled_units = [
        Yfarad,
        Zfarad,
        Efarad,
        Pfarad,
        Tfarad,
        Gfarad,
        Mfarad,
        kfarad,
        hfarad,
        dafarad,
        dfarad,
        cfarad,
        mfarad,
        ufarad,
        nfarad,
        pfarad,
        ffarad,
        afarad,
        zfarad,
        yfarad,
    ]

    # some powered units
    powered_units = [cmetre2, Yfarad3]

    # Combined units
    complex_units = [
        (kgram * metre2) / (amp * second3),
        5 * (kgram * metre2) / (amp * second3),
        metre * second**-1,
        10 * metre * second**-1,
        array([1, 2, 3]) * kmetre / second,
        np.ones(3) * nS / cm**2,
        # Made-up unit:
        Unit(
            1,
            dim=get_or_create_dimension(length=5, time=2),
            dispname="O",
            latexname=r"\Omega",
        ),
        8000 * umetre**3,
        [0.0001, 10000] * umetre**3,
        1 / metre,
        1 / (coulomb * metre**2),
        Unit(1) / second,
        3.0 * mM,
        5 * mole / liter,
        7 * liter / meter3,
        1 / second**2,
        volt**-2,
        (volt**2) ** -1,
        (1 / second) / meter,
        1 / (1 / second),
    ]

    unitless = [second / second, 5 * second / second, Unit(1)]

    for u in itertools.chain(
        units_which_should_exist,
        some_scaled_units,
        powered_units,
        complex_units,
        unitless,
    ):
        assert len(str(u)) > 0
        if not is_dimensionless(u):
            assert len(sympy.latex(u))
        assert get_dimensions(eval(repr(u))) == get_dimensions(u)
        assert_allclose(eval(repr(u)), u)

    for ar in [np.arange(10000) * mV, np.arange(100).reshape(10, 10) * mV]:
        latex_str = sympy.latex(ar)
        assert 0 < len(latex_str) < 1000  # arbitrary threshold, but see #1425

    # test the `DIMENSIONLESS` object
    assert str(DIMENSIONLESS) == "1"
    assert repr(DIMENSIONLESS) == "Dimension()"

    # test DimensionMismatchError (only that it works without raising an error
    for error in [
        DimensionMismatchError("A description"),
        DimensionMismatchError("A description", DIMENSIONLESS),
        DimensionMismatchError("A description", DIMENSIONLESS, second.dim),
    ]:
        assert len(str(error))
        assert len(repr(error))


@pytest.mark.codegen_independent
def test_format_quantity():
    # Avoid that the default f-string (or .format call) discards units when used without
    # a format spec
    q = 0.5 * ms
    assert f"{q}" == f"{q!s}" == str(q)
    assert f"{q:g}" == f"{float(q)}"


@pytest.mark.codegen_independent
def test_slicing():
    # Slicing and indexing, setting items
    quantity = np.reshape(np.arange(6), (2, 3)) * mV
    assert_equal(quantity[:], quantity)
    assert_equal(quantity[0], np.asarray(quantity)[0] * volt)
    assert_equal(quantity[0:1], np.asarray(quantity)[0:1] * volt)
    assert_equal(quantity[0, 1], np.asarray(quantity)[0, 1] * volt)
    assert_equal(quantity[0:1, 1:], np.asarray(quantity)[0:1, 1:] * volt)
    bool_matrix = np.array([[True, False, False], [False, False, True]])
    assert_equal(quantity[bool_matrix], np.asarray(quantity)[bool_matrix] * volt)


@pytest.mark.codegen_independent
def test_setting():
    quantity = np.reshape(np.arange(6), (2, 3)) * mV
    quantity[0, 1] = 10 * mV
    assert quantity[0, 1] == 10 * mV
    quantity[:, 1] = 20 * mV
    assert np.all(quantity[:, 1] == 20 * mV)
    quantity[1, :] = np.ones((1, 3)) * volt
    assert np.all(quantity[1, :] == 1 * volt)
    # Setting to zero should work without units as well
    quantity[1, 2] = 0
    assert quantity[1, 2] == 0 * mV

    def set_to_value(key, value):
        quantity[key] = value

    with pytest.raises(DimensionMismatchError):
        set_to_value(0, 1)
    with pytest.raises(DimensionMismatchError):
        set_to_value(0, 1 * second)
    with pytest.raises(DimensionMismatchError):
        set_to_value((slice(2), slice(3)), np.ones((2, 3)))


@pytest.mark.codegen_independent
def test_multiplication_division():
    quantities = [3 * mV, np.array([1, 2]) * mV, np.ones((3, 3)) * mV]
    q2 = 5 * second

    for q in quantities:
        # Scalars and array scalars
        assert_quantity(q / 3, np.asarray(q) / 3, volt)
        assert_quantity(3 / q, 3 / np.asarray(q), 1 / volt)
        assert_quantity(q * 3, np.asarray(q) * 3, volt)
        assert_quantity(3 * q, 3 * np.asarray(q), volt)
        assert_quantity(q / np.float64(3), np.asarray(q) / 3, volt)
        assert_quantity(np.float64(3) / q, 3 / np.asarray(q), 1 / volt)
        assert_quantity(q * np.float64(3), np.asarray(q) * 3, volt)
        assert_quantity(np.float64(3) * q, 3 * np.asarray(q), volt)
        assert_quantity(q / np.array(3), np.asarray(q) / 3, volt)
        assert_quantity(np.array(3) / q, 3 / np.asarray(q), 1 / volt)
        assert_quantity(q * np.array(3), np.asarray(q) * 3, volt)
        assert_quantity(np.array(3) * q, 3 * np.asarray(q), volt)

        # (unitless) arrays
        assert_quantity(q / np.array([3]), np.asarray(q) / 3, volt)
        assert_quantity(np.array([3]) / q, 3 / np.asarray(q), 1 / volt)
        assert_quantity(q * np.array([3]), np.asarray(q) * 3, volt)
        assert_quantity(np.array([3]) * q, 3 * np.asarray(q), volt)

        # arrays with units
        assert_quantity(q / q, np.asarray(q) / np.asarray(q), 1)
        assert_quantity(q * q, np.asarray(q) ** 2, volt**2)
        assert_quantity(q / q2, np.asarray(q) / np.asarray(q2), volt / second)
        assert_quantity(q2 / q, np.asarray(q2) / np.asarray(q), second / volt)
        assert_quantity(q * q2, np.asarray(q) * np.asarray(q2), volt * second)

        # using unsupported objects should fail
        with pytest.raises(TypeError):
            q / "string"
        with pytest.raises(TypeError):
            "string" / q
        with pytest.raises(TypeError):
            "string" * q
        with pytest.raises(TypeError):
            q * "string"


@pytest.mark.codegen_independent
def test_addition_subtraction():
    quantities = [3 * mV, np.array([1, 2]) * mV, np.ones((3, 3)) * mV]
    q2 = 5 * volt

    for q in quantities:
        # arrays with units
        assert_quantity(q + q, np.asarray(q) + np.asarray(q), volt)
        assert_quantity(q - q, 0, volt)
        assert_quantity(q + q2, np.asarray(q) + np.asarray(q2), volt)
        assert_quantity(q2 + q, np.asarray(q2) + np.asarray(q), volt)
        assert_quantity(q - q2, np.asarray(q) - np.asarray(q2), volt)
        assert_quantity(q2 - q, np.asarray(q2) - np.asarray(q), volt)

        # mismatching units
        with pytest.raises(DimensionMismatchError):
            q + 5 * second
        with pytest.raises(DimensionMismatchError):
            5 * second + q
        with pytest.raises(DimensionMismatchError):
            q - 5 * second
        with pytest.raises(DimensionMismatchError):
            5 * second - q

        # scalar
        with pytest.raises(DimensionMismatchError):
            q + 5
        with pytest.raises(DimensionMismatchError):
            5 + q
        with pytest.raises(DimensionMismatchError):
            q + np.float64(5)
        with pytest.raises(DimensionMismatchError):
            np.float64(5) + q
        with pytest.raises(DimensionMismatchError):
            q - 5
        with pytest.raises(DimensionMismatchError):
            5 - q
        with pytest.raises(DimensionMismatchError):
            q - np.float64(5)
        with pytest.raises(DimensionMismatchError):
            np.float64(5) - q

        # unitless array
        with pytest.raises(DimensionMismatchError):
            q + np.array([5])
        with pytest.raises(DimensionMismatchError):
            np.array([5]) + q
        with pytest.raises(DimensionMismatchError):
            q + np.array([5], dtype=np.float64)
        with pytest.raises(DimensionMismatchError):
            np.array([5], dtype=np.float64) + q
        with pytest.raises(DimensionMismatchError):
            q - np.array([5])
        with pytest.raises(DimensionMismatchError):
            np.array([5]) - q
        with pytest.raises(DimensionMismatchError):
            q - np.array([5], dtype=np.float64)
        with pytest.raises(DimensionMismatchError):
            np.array([5], dtype=np.float64) - q

        # Check that operations with 0 work
        assert_quantity(q + 0, np.asarray(q), volt)
        assert_quantity(0 + q, np.asarray(q), volt)
        assert_quantity(q - 0, np.asarray(q), volt)
        assert_quantity(0 - q, -np.asarray(q), volt)
        assert_quantity(q + np.float64(0), np.asarray(q), volt)
        assert_quantity(np.float64(0) + q, np.asarray(q), volt)
        assert_quantity(q - np.float64(0), np.asarray(q), volt)
        assert_quantity(np.float64(0) - q, -np.asarray(q), volt)

        # using unsupported objects should fail
        with pytest.raises(TypeError):
            "string" + q
        with pytest.raises(TypeError):
            q + "string"
        with pytest.raises(TypeError):
            q - "string"
        with pytest.raises(TypeError):
            "string" - q


@pytest.mark.codegen_independent
def test_unary_operations():
    from operator import neg, pos

    for op in [neg, pos]:
        for x in [2, np.array([2]), np.array([1, 2])]:
            assert_quantity(op(x * kilogram), op(x), kilogram)


@pytest.mark.codegen_independent
def test_binary_operations():
    """Test whether binary operations work when they should and raise
    DimensionMismatchErrors when they should.
    Does not test for the actual result.
    """
    from operator import add, eq, ge, gt, le, lt, ne, sub

    def assert_operations_work(a, b):
        try:
            # Test python builtins
            tryops = [add, sub, lt, le, gt, ge, eq, ne]
            for op in tryops:
                op(a, b)
                op(b, a)

            # Test equivalent numpy functions
            numpy_funcs = [
                np.add,
                np.subtract,
                np.less,
                np.less_equal,
                np.greater,
                np.greater_equal,
                np.equal,
                np.not_equal,
                np.maximum,
                np.minimum,
            ]
            for numpy_func in numpy_funcs:
                numpy_func(a, b)
                numpy_func(b, a)
        except DimensionMismatchError as ex:
            raise AssertionError(f"Operation raised unexpected exception: {ex}")

    def assert_operations_do_not_work(a, b):
        # Test python builtins
        tryops = [add, sub, lt, le, gt, ge, eq, ne]
        for op in tryops:
            with pytest.raises(DimensionMismatchError):
                op(a, b)
            with pytest.raises(DimensionMismatchError):
                op(b, a)

        # Test equivalent numpy functions
        numpy_funcs = [
            np.add,
            np.subtract,
            np.less,
            np.less_equal,
            np.greater,
            np.greater_equal,
            np.equal,
            np.not_equal,
            np.maximum,
            np.minimum,
        ]
        for numpy_func in numpy_funcs:
            with pytest.raises(DimensionMismatchError):
                numpy_func(a, b)
            with pytest.raises(DimensionMismatchError):
                numpy_func(b, a)

    #
    # Check that consistent units work
    #

    # unit arrays
    a = 1 * kilogram
    for b in [2 * kilogram, np.array([2]) * kilogram, np.array([1, 2]) * kilogram]:
        assert_operations_work(a, b)

    # dimensionless units and scalars
    a = 1
    for b in [
        2 * kilogram / kilogram,
        np.array([2]) * kilogram / kilogram,
        np.array([1, 2]) * kilogram / kilogram,
    ]:
        assert_operations_work(a, b)

    # dimensionless units and unitless arrays
    a = np.array([1])
    for b in [
        2 * kilogram / kilogram,
        np.array([2]) * kilogram / kilogram,
        np.array([1, 2]) * kilogram / kilogram,
    ]:
        assert_operations_work(a, b)

    #
    # Check that inconsistent units do not work
    #

    # unit arrays
    a = np.array([1]) * second
    for b in [2 * kilogram, np.array([2]) * kilogram, np.array([1, 2]) * kilogram]:
        assert_operations_do_not_work(a, b)

    # unitless array
    a = np.array([1])
    for b in [2 * kilogram, np.array([2]) * kilogram, np.array([1, 2]) * kilogram]:
        assert_operations_do_not_work(a, b)

    # scalar
    a = 1
    for b in [2 * kilogram, np.array([2]) * kilogram, np.array([1, 2]) * kilogram]:
        assert_operations_do_not_work(a, b)

    # Check that comparisons with inf/-inf always work
    values = [
        2 * kilogram / kilogram,
        2 * kilogram,
        np.array([2]) * kilogram,
        np.array([1, 2]) * kilogram,
    ]
    for value in values:
        assert np.all(value < np.inf)
        assert np.all(np.inf > value)
        assert np.all(value <= np.inf)
        assert np.all(np.inf >= value)
        assert np.all(value != np.inf)
        assert np.all(np.inf != value)
        assert np.all(value >= -np.inf)
        assert np.all(-np.inf <= value)
        assert np.all(value > -np.inf)
        assert np.all(-np.inf < value)


@pytest.mark.codegen_independent
def test_power():
    """
    Test raising quantities to a power.
    """
    values = [2 * kilogram, np.array([2]) * kilogram, np.array([1, 2]) * kilogram]
    for value in values:
        assert_quantity(value**3, np.asarray(value) ** 3, kilogram**3)
        # Test raising to a dimensionless quantity
        assert_quantity(value ** (3 * volt / volt), np.asarray(value) ** 3, kilogram**3)
        with pytest.raises(DimensionMismatchError):
            # FIXME: Not that if float(exponent) is a special value such as 1 or 2
            # numpy will actually use a ufunc such as identity or square, which will
            # not raise a DimensionMismatchError. This is a limitation of the current
            # implementation.
            value ** (2 * mV)
        with pytest.raises(TypeError):
            value ** np.array([2, 3])


@pytest.mark.codegen_independent
def test_inplace_operations():
    q = np.arange(10) * volt
    q_orig = q.copy()
    q_ref = q

    q *= 2
    assert np.array_equal(q, 2 * q_orig)
    assert np.array_equal(q_ref, q)
    q /= 2
    assert np.array_equal(q, q_orig)
    assert np.array_equal(q_ref, q)
    q += 1 * volt
    assert np.array_equal(q, q_orig + 1 * volt)
    assert np.array_equal(q_ref, q)
    q -= 1 * volt
    assert np.array_equal(q, q_orig)
    assert np.array_equal(q_ref, q)
    q **= 2
    assert np.array_equal(q, q_orig**2)
    assert np.array_equal(q_ref, q)
    q **= 0.5
    assert np.array_equal(q, q_orig)
    assert np.array_equal(q_ref, q)

    def illegal_add(q2):
        q = np.arange(10) * volt
        q += q2

    with pytest.raises(DimensionMismatchError):
        illegal_add(1 * second)
    with pytest.raises(DimensionMismatchError):
        illegal_add(1)

    def illegal_sub(q2):
        q = np.arange(10) * volt
        q -= q2

    with pytest.raises(DimensionMismatchError):
        illegal_add(1 * second)
    with pytest.raises(DimensionMismatchError):
        illegal_add(1)

    def illegal_pow(q2):
        q = np.arange(10) * volt
        q **= q2

    with pytest.raises(DimensionMismatchError):
        illegal_pow(1 * mV)
    with pytest.raises(TypeError):
        illegal_pow(np.arange(10))

    # inplace operations with unsupported objects should fail
    for inplace_op in [
        q.__iadd__,
        q.__isub__,
        q.__imul__,
        q.__itruediv__,
        q.__ifloordiv__,
        q.__imod__,
        q.__ipow__,
    ]:
        try:
            result = inplace_op("string")
            # if it doesn't fail with an error, it should return NotImplemented
            assert result == NotImplemented
        except TypeError:
            pass  # raised on numpy >= 0.10

    # make sure that inplace operations do not work on units/dimensions at all
    for inplace_op in [
        volt.__iadd__,
        volt.__isub__,
        volt.__imul__,
        volt.__itruediv__,
        volt.__ifloordiv__,
        volt.__imod__,
        volt.__ipow__,
    ]:
        with pytest.raises(TypeError):
            inplace_op(volt)
    for inplace_op in [
        volt.dimensions.__imul__,
        volt.dimensions.__itruediv__,
        volt.dimensions.__ipow__,
    ]:
        with pytest.raises(TypeError):
            inplace_op(volt.dimensions)


@pytest.mark.codegen_independent
def test_unit_discarding_functions():
    """
    Test functions that discard units.
    """
    from brian2.units.unitsafefunctions import ones_like, zeros_like

    values = [3 * mV, np.array([1, 2]) * mV, np.arange(12).reshape(3, 4) * mV]
    for value in values:
        assert_equal(np.sign(value), np.sign(np.asarray(value)))
        assert_equal(zeros_like(value), np.zeros_like(np.asarray(value)))
        assert_equal(ones_like(value), np.ones_like(np.asarray(value)))
        # Calling non-zero on a 0d array is deprecated, don't test it:
        if value.ndim > 0:
            assert_equal(np.nonzero(value), np.nonzero(np.asarray(value)))


@pytest.mark.codegen_independent
def test_unitsafe_functions():
    """
    Test the unitsafe functions wrapping their numpy counterparts.
    """
    from brian2.units.unitsafefunctions import (
        arccos,
        arccosh,
        arcsin,
        arcsinh,
        arctan,
        arctanh,
        cos,
        cosh,
        exp,
        log,
        sin,
        sinh,
        tan,
        tanh,
    )

    # All functions with their numpy counterparts
    funcs = [
        (sin, np.sin),
        (sinh, np.sinh),
        (arcsin, np.arcsin),
        (arcsinh, np.arcsinh),
        (cos, np.cos),
        (cosh, np.cosh),
        (arccos, np.arccos),
        (arccosh, np.arccosh),
        (tan, np.tan),
        (tanh, np.tanh),
        (arctan, np.arctan),
        (arctanh, np.arctanh),
        (log, np.log),
        (exp, np.exp),
    ]

    unitless_values = [
        3 * mV / mV,
        np.array([1, 2]) * mV / mV,
        np.ones((3, 3)) * mV / mV,
    ]
    numpy_values = [3, np.array([1, 2]), np.ones((3, 3))]
    unit_values = [3 * mV, np.array([1, 2]) * mV, np.ones((3, 3)) * mV]

    for func, np_func in funcs:
        # make sure these functions raise errors when run on values with dimensions
        for val in unit_values:
            with pytest.raises(DimensionMismatchError):
                func(val)

        # make sure the functions are equivalent to their numpy counterparts
        # when run on unitless values while ignoring warnings about invalid
        # values or divisions by zero
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            for val in unitless_values:
                assert_equal(func(val), np_func(val))

            for val in numpy_values:
                assert_equal(func(val), np_func(val))


@pytest.mark.codegen_independent
def test_special_case_numpy_functions():
    """
    Test a couple of functions/methods that need special treatment.
    """
    from brian2.units.unitsafefunctions import diagonal, dot, ravel, trace, where

    quadratic_matrix = np.reshape(np.arange(9), (3, 3)) * mV

    # Temporarily suppress warnings related to the matplotlib 1.3 bug
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # Check that function and method do the same thing
        assert_equal(ravel(quadratic_matrix), quadratic_matrix.ravel())
        # Check that function gives the same result as on unitless arrays
        assert_equal(
            np.asarray(ravel(quadratic_matrix)), ravel(np.asarray(quadratic_matrix))
        )
        # Check that the function gives the same results as the original numpy
        # function
        assert_equal(
            np.ravel(np.asarray(quadratic_matrix)), ravel(np.asarray(quadratic_matrix))
        )

    # Do the same checks for diagonal, trace and dot
    assert_equal(diagonal(quadratic_matrix), quadratic_matrix.diagonal())
    assert_equal(
        np.asarray(diagonal(quadratic_matrix)), diagonal(np.asarray(quadratic_matrix))
    )
    assert_equal(
        np.diagonal(np.asarray(quadratic_matrix)),
        diagonal(np.asarray(quadratic_matrix)),
    )

    assert_equal(trace(quadratic_matrix), quadratic_matrix.trace())
    assert_equal(
        np.asarray(trace(quadratic_matrix)), trace(np.asarray(quadratic_matrix))
    )
    assert_equal(
        np.trace(np.asarray(quadratic_matrix)), trace(np.asarray(quadratic_matrix))
    )

    assert_equal(
        dot(quadratic_matrix, quadratic_matrix), quadratic_matrix.dot(quadratic_matrix)
    )
    assert_equal(
        np.asarray(dot(quadratic_matrix, quadratic_matrix)),
        dot(np.asarray(quadratic_matrix), np.asarray(quadratic_matrix)),
    )
    assert_equal(
        np.dot(np.asarray(quadratic_matrix), np.asarray(quadratic_matrix)),
        dot(np.asarray(quadratic_matrix), np.asarray(quadratic_matrix)),
    )

    assert_equal(
        np.asarray(quadratic_matrix.prod()), np.asarray(quadratic_matrix).prod()
    )
    assert_equal(
        np.asarray(quadratic_matrix.prod(axis=0)),
        np.asarray(quadratic_matrix).prod(axis=0),
    )

    # Check for correct units
    if use_matplotlib_units_fix:
        assert have_same_dimensions(1, ravel(quadratic_matrix))
    else:
        assert have_same_dimensions(quadratic_matrix, ravel(quadratic_matrix))
    assert have_same_dimensions(quadratic_matrix, trace(quadratic_matrix))
    assert have_same_dimensions(quadratic_matrix, diagonal(quadratic_matrix))
    assert have_same_dimensions(
        quadratic_matrix[0] ** 2, dot(quadratic_matrix, quadratic_matrix)
    )
    assert have_same_dimensions(
        quadratic_matrix.prod(axis=0), quadratic_matrix[0] ** quadratic_matrix.shape[0]
    )

    # check the where function
    # pure numpy array
    cond = [True, False, False]
    ar1 = np.array([1, 2, 3])
    ar2 = np.array([4, 5, 6])
    assert_equal(np.where(cond), where(cond))
    assert_equal(np.where(cond, ar1, ar2), where(cond, ar1, ar2))

    # dimensionless quantity
    assert_equal(
        np.where(cond, ar1, ar2), np.asarray(where(cond, ar1 * mV / mV, ar2 * mV / mV))
    )

    # quantity with dimensions
    ar1 = ar1 * mV
    ar2 = ar2 * mV
    assert_equal(
        np.where(cond, np.asarray(ar1), np.asarray(ar2)),
        np.asarray(where(cond, ar1, ar2)),
    )

    # Check some error cases
    with pytest.raises(ValueError):
        where(cond, ar1)
    with pytest.raises(TypeError):
        where(cond, ar1, ar1, ar2)
    with pytest.raises(DimensionMismatchError):
        where(cond, ar1, ar1 / ms)

    # Check setasflat (for numpy < 1.7)
    if hasattr(Quantity, "setasflat"):
        a = np.arange(10) * mV
        b = np.ones(10).reshape(5, 2) * volt
        c = np.ones(10).reshape(5, 2) * second
        with pytest.raises(DimensionMismatchError):
            a.setasflat(c)
        a.setasflat(b)
        assert_equal(a.flatten(), b.flatten())

    # Check cumprod
    a = np.arange(1, 10) * mV / mV
    assert_equal(a.cumprod(), np.asarray(a).cumprod())
    with pytest.raises(TypeError):
        (np.arange(1, 5) * mV).cumprod()


# Functions that should not change units
@pytest.mark.codegen_independent
def test_numpy_functions_same_dimensions():
    values = [np.array([1, 2]), np.ones((3, 3))]
    units = [volt, second, siemens, mV, kHz]

    from brian2.units.unitsafefunctions import ptp

    # numpy functions
    keep_dim_funcs = [
        np.abs,
        np.cumsum,
        np.max,
        np.mean,
        np.min,
        np.negative,
        ptp,
        np.round,
        np.squeeze,
        np.std,
        np.sum,
        np.transpose,
    ]

    for value, unit in itertools.product(values, units):
        q_ar = value * unit
        for func in keep_dim_funcs:
            test_ar = func(q_ar)
            if not get_dimensions(test_ar) is q_ar.dim:
                raise AssertionError(
                    f"'{func.__name__}' failed on {q_ar!r} -- dim was "
                    f"{q_ar.dim}, is now {get_dimensions(test_ar)}."
                )

                # Python builtins should work on one-dimensional arrays
                value = np.arange(5)
                builtins = [abs, max, min, sum]
                for unit in units:
                    q_ar = value * unit
                for func in builtins:
                    test_ar = func(q_ar)
                if not get_dimensions(test_ar) is q_ar.dim:
                    raise AssertionError(
                        f"'{func.__name__}' failed on {q_ar!r} -- dim "
                        f"was {q_ar.dim}, is now "
                        f"{get_dimensions(test_ar)}"
                    )


@pytest.mark.codegen_independent
def test_numpy_functions_indices():
    """
    Check numpy functions that return indices.
    """
    values = [np.array([-4, 3, -2, 1, 0]), np.ones((3, 3)), np.array([17])]
    units = [volt, second, siemens, mV, kHz]

    # numpy functions
    keep_dim_funcs = [np.argmin, np.argmax, np.argsort, np.nonzero]

    for value, unit in itertools.product(values, units):
        q_ar = value * unit
        for func in keep_dim_funcs:
            test_ar = func(q_ar)
            # Compare it to the result on the same value without units
            comparison_ar = func(value)
            assert_equal(
                test_ar,
                comparison_ar,
                (
                    "function %s returned an incorrect result when used on quantities "
                    % func.__name__
                ),
            )


@pytest.mark.codegen_independent
def test_numpy_functions_dimensionless():
    """
    Test that numpy functions that should work on dimensionless quantities only
    work dimensionless arrays and return the correct result.
    """
    unitless_values = [3, np.array([-4, 3, -1, 2]), np.ones((3, 3))]
    unit_values = [3 * mV, np.array([-4, 3, -1, 2]) * mV, np.ones((3, 3)) * mV]
    with warnings.catch_warnings():
        # ignore division by 0 warnings
        warnings.simplefilter("ignore", RuntimeWarning)
        for value in unitless_values:
            for ufunc in UFUNCS_DIMENSIONLESS:
                result_unitless = eval(f"np.{ufunc}(value)")
                result_array = eval(f"np.{ufunc}(np.array(value))")
                assert isinstance(
                    result_unitless, (np.ndarray, np.number)
                ) and not isinstance(result_unitless, Quantity)
                assert_equal(result_unitless, result_array)
            for ufunc in UFUNCS_DIMENSIONLESS_TWOARGS:
                result_unitless = eval(f"np.{ufunc}(value, value)")
                result_array = eval(f"np.{ufunc}(np.array(value), np.array(value))")
                assert isinstance(
                    result_unitless, (np.ndarray, np.number)
                ) and not isinstance(result_unitless, Quantity)
                assert_equal(result_unitless, result_array)

        for value, unitless_value in zip(unit_values, unitless_values):
            for ufunc in UFUNCS_DIMENSIONLESS:
                with pytest.raises(DimensionMismatchError):
                    eval(f"np.{ufunc}(value)", globals(), {"value": value})
            for ufunc in UFUNCS_DIMENSIONLESS_TWOARGS:
                with pytest.raises(DimensionMismatchError):
                    eval(
                        f"np.{ufunc}(value1, value2)",
                        globals(),
                        {"value1": value, "value2": unitless_value},
                    )
                with pytest.raises(DimensionMismatchError):
                    eval(
                        f"np.{ufunc}(value2, value1)",
                        globals(),
                        {"value1": value, "value2": unitless_value},
                    )
                with pytest.raises(DimensionMismatchError):
                    eval(f"np.{ufunc}(value, value)", globals(), {"value": value})


@pytest.mark.codegen_independent
def test_numpy_functions_change_dimensions():
    """
    Test some numpy functions that change the dimensions of the quantity.
    """
    unit_values = [np.array([1, 2]) * mV, np.ones((3, 3)) * 2 * mV]
    for value in unit_values:
        assert_quantity(np.var(value), np.var(np.array(value)), volt**2)
        assert_quantity(np.square(value), np.square(np.array(value)), volt**2)
        assert_quantity(np.sqrt(value), np.sqrt(np.array(value)), volt**0.5)
        assert_quantity(
            np.reciprocal(value), np.reciprocal(np.array(value)), 1.0 / volt
        )


@pytest.mark.codegen_independent
def test_numpy_functions_matmul():
    """
    Check support for matmul and the ``@`` operator.
    """
    no_units_eye = np.eye(3)
    with_units_eye = no_units_eye * Mohm
    matrix_no_units = np.arange(9).reshape((3, 3))
    matrix_units = matrix_no_units * nA

    # First operand with units
    assert_allclose(no_units_eye @ matrix_units, matrix_units)
    assert have_same_dimensions(no_units_eye @ matrix_units, matrix_units)
    assert_allclose(np.matmul(no_units_eye, matrix_units), matrix_units)
    assert have_same_dimensions(np.matmul(no_units_eye, matrix_units), matrix_units)

    # Second operand with units
    assert_allclose(with_units_eye @ matrix_no_units, matrix_no_units * Mohm)
    assert have_same_dimensions(
        with_units_eye @ matrix_no_units, matrix_no_units * Mohm
    )
    assert_allclose(np.matmul(with_units_eye, matrix_no_units), matrix_no_units * Mohm)
    assert have_same_dimensions(
        np.matmul(with_units_eye, matrix_no_units), matrix_no_units * Mohm
    )

    # Both operands with units
    assert_allclose(
        with_units_eye @ matrix_units, no_units_eye @ matrix_no_units * nA * Mohm
    )
    assert have_same_dimensions(with_units_eye @ matrix_units, nA * Mohm)
    assert_allclose(
        np.matmul(with_units_eye, matrix_units),
        np.matmul(no_units_eye, matrix_no_units) * nA * Mohm,
    )
    assert have_same_dimensions(np.matmul(with_units_eye, matrix_units), nA * Mohm)


@pytest.mark.codegen_independent
def test_numpy_functions_typeerror():
    """
    Assures that certain numpy functions raise a TypeError when called on
    quantities.
    """
    unitless_values = [
        3 * mV / mV,
        np.array([1, 2]) * mV / mV,
        np.ones((3, 3)) * mV / mV,
    ]
    unit_values = [3 * mV, np.array([1, 2]) * mV, np.ones((3, 3)) * mV]
    for value in unitless_values + unit_values:
        for ufunc in UFUNCS_INTEGERS:
            if ufunc == "invert":
                # only takes one argument
                with pytest.raises(TypeError):
                    eval(f"np.{ufunc}(value)", globals(), {"value": value})
            else:
                with pytest.raises(TypeError):
                    eval(f"np.{ufunc}(value, value)", globals(), {"value": value})


@pytest.mark.codegen_independent
def test_numpy_functions_logical():
    """
    Assure that logical numpy functions work on all quantities and return
    unitless boolean arrays.
    """
    unit_values1 = [3 * mV, np.array([1, 2]) * mV, np.ones((3, 3)) * mV]
    unit_values2 = [3 * second, np.array([1, 2]) * second, np.ones((3, 3)) * second]
    for ufunc in UFUNCS_LOGICAL:
        for value1, value2 in zip(unit_values1, unit_values2):
            try:
                # one argument
                result_units = eval(f"np.{ufunc}(value1)")
                result_array = eval(f"np.{ufunc}(np.array(value1))")
            except (ValueError, TypeError):
                # two arguments
                result_units = eval(f"np.{ufunc}(value1, value2)")
                result_array = eval(f"np.{ufunc}(np.array(value1), np.array(value2))")
            assert not isinstance(result_units, Quantity)
            assert_equal(result_units, result_array)


@pytest.mark.codegen_independent
def test_arange_linspace():
    # For dimensionless values, the unit-safe functions should give the same results
    assert_equal(brian2.arange(5), np.arange(5))
    assert_equal(brian2.arange(1, 5), np.arange(1, 5))
    assert_equal(brian2.arange(10, step=2), np.arange(10, step=2))
    assert_equal(brian2.arange(0, 5, 0.5), np.arange(0, 5, 0.5))
    assert_equal(brian2.linspace(0, 1), np.linspace(0, 1))
    assert_equal(brian2.linspace(0, 1, 10), np.linspace(0, 1, 10))

    # Make sure units are checked
    with pytest.raises(DimensionMismatchError):
        brian2.arange(1 * mV, 5)
    with pytest.raises(DimensionMismatchError):
        brian2.arange(1 * mV, 5 * mV)
    with pytest.raises(DimensionMismatchError):
        brian2.arange(1, 5 * mV)
    with pytest.raises(DimensionMismatchError):
        brian2.arange(1 * mV, 5 * ms)
    with pytest.raises(DimensionMismatchError):
        brian2.arange(1 * mV, 5 * mV, step=1 * ms)
    with pytest.raises(DimensionMismatchError):
        brian2.arange(1 * ms, 5 * mV)

    # Check correct functioning with units
    assert_quantity(
        brian2.arange(5 * mV, step=1 * mV), float(mV) * np.arange(5, step=1), mV
    )
    assert_quantity(
        brian2.arange(1 * mV, 5 * mV, 1 * mV), float(mV) * np.arange(1, 5, 1), mV
    )
    assert_quantity(brian2.linspace(1 * mV, 2 * mV), float(mV) * np.linspace(1, 2), mV)

    # Check errors for arange with incorrect numbers of arguments/duplicate arguments
    with pytest.raises(TypeError):
        brian2.arange()
    with pytest.raises(TypeError):
        brian2.arange(0, 5, 1, 0)
    with pytest.raises(TypeError):
        brian2.arange(0, stop=1)
    with pytest.raises(TypeError):
        brian2.arange(0, 5, stop=1)
    with pytest.raises(TypeError):
        brian2.arange(0, 5, start=1)
    with pytest.raises(TypeError):
        brian2.arange(0, 5, 1, start=1)
    with pytest.raises(TypeError):
        brian2.arange(0, 5, 1, stop=2)
    with pytest.raises(TypeError):
        brian2.arange(0, 5, 1, step=2)


@pytest.mark.codegen_independent
def test_list():
    """
    Test converting to and from a list.
    """
    values = [3 * mV, np.array([1, 2]) * mV, np.arange(12).reshape(4, 3) * mV]
    for value in values:
        l = value.tolist()
        from_list = Quantity(l)
        assert have_same_dimensions(from_list, value)
        assert_equal(from_list, value)


@pytest.mark.codegen_independent
def test_check_units():
    """
    Test the check_units decorator
    """

    @check_units(v=volt)
    def a_function(v, x):
        """
        v has to have units of volt, x can have any (or no) unit.
        """
        pass

    # Try correct units
    a_function(3 * mV, 5 * second)
    a_function(5 * volt, "something")
    a_function([1, 2, 3] * volt, None)
    # lists that can be converted should also work
    a_function([1 * volt, 2 * volt, 3 * volt], None)
    # Strings and None are also allowed to pass
    a_function("a string", None)
    a_function(None, None)

    # Try incorrect units
    with pytest.raises(DimensionMismatchError):
        a_function(5 * second, None)
    with pytest.raises(DimensionMismatchError):
        a_function(5, None)
    with pytest.raises(TypeError):
        a_function(object(), None)
    with pytest.raises(TypeError):
        a_function([1, 2 * volt, 3], None)

    @check_units(result=second)
    def b_function(return_second):
        """
        Return a value in seconds if return_second is True, otherwise return
        a value in volt.
        """
        if return_second:
            return 5 * second
        else:
            return 3 * volt

    # Should work (returns second)
    b_function(True)
    # Should fail (returns volt)
    with pytest.raises(DimensionMismatchError):
        b_function(False)

    @check_units(a=bool, b=1, result=bool)
    def c_function(a, b):
        if a:
            return b > 0
        else:
            return b

    assert c_function(True, 1)
    assert not c_function(True, -1)
    with pytest.raises(TypeError):
        c_function(1, 1)
    with pytest.raises(TypeError):
        c_function(1 * mV, 1)
    with pytest.raises(TypeError):
        c_function(False, 1)


@pytest.mark.codegen_independent
def test_get_unit():
    """
    Test get_unit
    """
    values = [
        (volt.dim, volt),
        (mV.dim, volt),
        ((amp / metre**2).dim, amp / metre**2),
    ]
    for dim, expected_unit in values:
        unit = get_unit(dim)
        assert isinstance(unit, Unit)
        assert unit == expected_unit
        assert float(unit) == 1.0


@pytest.mark.codegen_independent
def test_get_best_unit():
    # get_best_unit should not check all values for long arrays, since it is
    # a function used for display purposes only. Instead, only the first and
    # last few values should matter (see github issue #966)
    long_ar = np.ones(10000) * siemens
    long_ar[:10] = 1 * nS
    long_ar[-10:] = 2 * nS
    values = [
        (np.arange(10) * mV, mV),
        ([0.001, 0.002, 0.003] * second, ms),
        (long_ar, nS),
    ]
    for ar, expected_unit in values:
        assert ar.get_best_unit() is expected_unit
        assert str(expected_unit) in ar.in_best_unit()


@pytest.mark.codegen_independent
def test_switching_off_unit_checks():
    """
    Check switching off unit checks (used for external functions).
    """
    import brian2.units.fundamentalunits as fundamentalunits

    x = 3 * second
    y = 5 * volt
    with pytest.raises(DimensionMismatchError):
        x + y
    fundamentalunits.unit_checking = False
    # Now it should work
    assert np.asarray(x + y) == np.array(8)
    assert have_same_dimensions(x, y)
    assert x.has_same_dimensions(y)
    fundamentalunits.unit_checking = True


@pytest.mark.codegen_independent
def test_fail_for_dimension_mismatch():
    """
    Test the fail_for_dimension_mismatch function.
    """
    # examples that should not raise an error
    dim1, dim2 = fail_for_dimension_mismatch(3)
    assert dim1 is DIMENSIONLESS
    assert dim2 is DIMENSIONLESS
    dim1, dim2 = fail_for_dimension_mismatch(3 * volt / volt)
    assert dim1 is DIMENSIONLESS
    assert dim2 is DIMENSIONLESS
    dim1, dim2 = fail_for_dimension_mismatch(3 * volt / volt, 7)
    assert dim1 is DIMENSIONLESS
    assert dim2 is DIMENSIONLESS
    dim1, dim2 = fail_for_dimension_mismatch(3 * volt, 5 * volt)
    assert dim1 is volt.dim
    assert dim2 is volt.dim

    # examples that should raise an error
    with pytest.raises(DimensionMismatchError):
        fail_for_dimension_mismatch(6 * volt)
    with pytest.raises(DimensionMismatchError):
        fail_for_dimension_mismatch(6 * volt, 5 * second)


@pytest.mark.codegen_independent
def test_deepcopy():
    d = {"x": 1 * second}
    from copy import deepcopy

    d_copy = deepcopy(d)
    assert d_copy["x"] == 1 * second
    d_copy["x"] += 1 * second
    assert d_copy["x"] == 2 * second
    assert d["x"] == 1 * second


@pytest.mark.codegen_independent
def test_inplace_on_scalars():
    # We want "copy semantics" for in-place operations on scalar quantities
    # in the same way as for Python scalars
    for scalar in [3 * mV, 3 * mV / mV]:
        scalar_reference = scalar
        scalar_copy = Quantity(scalar, copy=True)
        scalar += scalar_copy
        assert_equal(scalar_copy, scalar_reference)
        scalar *= 1.5
        assert_equal(scalar_copy, scalar_reference)
        scalar /= 2
        assert_equal(scalar_copy, scalar_reference)

        # also check that it worked correctly for the scalar itself
        assert_allclose(scalar, (scalar_copy + scalar_copy) * 1.5 / 2)

    # For arrays, it should use reference semantics
    for vector in [[3] * mV, [3] * mV / mV]:
        vector_reference = vector
        vector_copy = Quantity(vector, copy=True)
        vector += vector_copy
        assert_equal(vector, vector_reference)
        vector *= 1.5
        assert_equal(vector, vector_reference)
        vector /= 2
        assert_equal(vector, vector_reference)

        # also check that it worked correctly for the vector itself
        assert_allclose(vector, (vector_copy + vector_copy) * 1.5 / 2)


def test_units_vs_quantities():
    # Unit objects should stay Unit objects under certain operations
    # (important e.g. in the unit definition of Equations, where only units but
    # not quantities are allowed)
    assert isinstance(meter**2, Unit)
    assert isinstance(meter**-1, Unit)
    assert isinstance(meter**0.5, Unit)
    assert isinstance(meter / second, Unit)
    assert isinstance(amp / meter**2, Unit)
    assert isinstance(1 / meter, Unit)
    assert isinstance(1.0 / meter, Unit)

    # Using the unconventional type(x) == y since we want to test that
    # e.g. meter**2 stays a Unit and does not become a Quantity however Unit
    # inherits from Quantity and therefore both would pass the isinstance test
    assert type(2 / meter) == Quantity
    assert type(2 * meter) == Quantity
    assert type(meter + meter) == Quantity
    assert type(meter - meter) == Quantity


@pytest.mark.codegen_independent
def test_all_units_list():
    from brian2.units.allunits import all_units

    assert meter in all_units
    assert volt in all_units
    assert cm in all_units
    assert Hz in all_units
    assert all(isinstance(u, Unit) for u in all_units)


@pytest.mark.codegen_independent
def test_constants():
    import brian2.units.constants as constants

    # Check that the expected names exist and have the correct dimensions
    assert constants.avogadro_constant.dim == (1 / mole).dim
    assert constants.boltzmann_constant.dim == (joule / kelvin).dim
    assert constants.electric_constant.dim == (farad / meter).dim
    assert constants.electron_mass.dim == kilogram.dim
    assert constants.elementary_charge.dim == coulomb.dim
    assert constants.faraday_constant.dim == (coulomb / mole).dim
    assert constants.gas_constant.dim == (joule / mole / kelvin).dim
    assert constants.magnetic_constant.dim == (newton / amp2).dim
    assert constants.molar_mass_constant.dim == (kilogram / mole).dim
    assert constants.zero_celsius.dim == kelvin.dim

    # Check the consistency between a few constants
    assert_allclose(
        constants.gas_constant,
        constants.avogadro_constant * constants.boltzmann_constant,
    )
    assert_allclose(
        constants.faraday_constant,
        constants.avogadro_constant * constants.elementary_charge,
    )


if __name__ == "__main__":
    test_construction()
    test_get_dimensions()
    test_display()
    test_scale()
    test_power()
    test_pickling()
    test_str_repr()
    test_slicing()
    test_setting()
    test_multiplication_division()
    test_addition_subtraction()
    test_unary_operations()
    test_binary_operations()
    test_inplace_operations()
    test_unit_discarding_functions()
    test_unitsafe_functions()
    test_special_case_numpy_functions()
    test_numpy_functions_same_dimensions()
    test_numpy_functions_indices()
    test_numpy_functions_dimensionless()
    test_numpy_functions_change_dimensions()
    test_numpy_functions_typeerror()
    test_numpy_functions_logical()
    test_arange_linspace()
    test_list()
    test_check_units()
    test_get_unit()
    test_get_best_unit()
    test_switching_off_unit_checks()
    test_fail_for_dimension_mismatch()
    test_deepcopy()
    test_inplace_on_scalars()
    test_units_vs_quantities()
    test_all_units_list()
    test_constants()



"""
An (almost) infinite cable with pulse injection in the middle.
"""
from brian2 import *

defaultclock.dt = 0.001*ms

# Morphology
diameter = 1*um
Cm = 1*uF/cm**2
Ri = 100*ohm*cm
N = 500
morpho = Cylinder(diameter=diameter, length=3*mm, n=N)

# Passive channels
gL = 1e-4*siemens/cm**2
EL = -70*mV
eqs = '''
Im = gL * (EL-v) : amp/meter**2
I : amp (point current)
'''

neuron = SpatialNeuron(morphology=morpho, model=eqs, Cm=Cm, Ri=Ri,
                       method = 'exponential_euler')
neuron.v = EL

taum = Cm  /gL  # membrane time constant
print("Time constant: %s" % taum)
la = neuron.space_constant[0]
print("Characteristic length: %s" % la)

# Monitors
mon = StateMonitor(neuron, 'v', record=range(0, N//2, 20))

neuron.I[len(neuron) // 2] = 1*nA  # injecting in the middle
run(0.02*ms)
neuron.I = 0*amp
run(10*ms, report='text')

t = mon.t
plot(t/ms, mon.v.T/mV, 'k')
# Theory (incorrect near cable ends)
for i in range(0, len(neuron)//2, 20):
    x = (len(neuron)/2 - i) * morpho.length[0]
    theory = (1/(la*Cm*pi*diameter) * sqrt(taum / (4*pi*(t + defaultclock.dt))) *
              exp(-(t+defaultclock.dt)/taum -
                  taum / (4*(t+defaultclock.dt))*(x/la)**2))
    theory = EL + theory * 1*nA * 0.02*ms
    plot(t/ms, theory/mV, 'r')
xlabel('Time (ms)')
ylabel('v (mV')
show()



#!/usr/bin/env python
"""
Reliability of spike timing.

See e.g. Mainen & Sejnowski (1995) for experimental results in vitro.
"""
from brian2 import *

# The common noisy input
N = 25
tau_input = 5*ms
neuron_input = NeuronGroup(1, 'dx/dt = -x / tau_input + (2 /tau_input)**.5 * xi : 1')

# The noisy neurons receiving the same input
tau = 10*ms
sigma = .015
eqs_neurons = '''
dx/dt = (0.9 + .5 * I - x) / tau + sigma * (2 / tau)**.5 * xi : 1
I : 1 (linked)
'''
neurons = NeuronGroup(N, model=eqs_neurons, threshold='x > 1',
                      reset='x = 0', refractory=5*ms, method='euler')
neurons.x = 'rand()'
neurons.I = linked_var(neuron_input, 'x') # input.x is continuously fed into neurons.I
spikes = SpikeMonitor(neurons)

run(500*ms)
plt.plot(spikes.t/ms, spikes.i, '.k')
xlabel('Time (ms)')
ylabel('Neuron index')
show()



{# USES_VARIABLES { N, _indices } #}
{% extends 'common_group.pyx' %}

{% block maincode %}

    _vectorisation_idx = 1

    cdef size_t _num_elements = 0
    cdef _numpy.ndarray[int, ndim=1, mode='c'] _elements = _numpy.zeros(N, dtype=_numpy.int32)
    cdef int[:] _elements_view = _elements
    
    {{scalar_code|autoindent}}

    for _idx in range(N):
        _vectorisation_idx = _idx
        
        {{vector_code|autoindent}}
        
        if _cond:
            _elements_view[_num_elements] = _idx
            _num_elements += 1
    
    return _elements[:_num_elements]

{% endblock %}
